from acoustic_toolbox.directivity import figure_eight
import numpy as np
import pytest


class TestDirectivity:
    """Test :mod:`acoustic_toolbox.directivity`"""

    @pytest.mark.parametrize(
        "given, expected, uncertainty",
        [
            (0.0, 1.0, 0.0),
            (1.0 / 2.0 * np.pi, 0.0, 0.0),
            (np.pi, +1.0, 0.0),
            (3.0 / 2.0 * np.pi, 0.0, 0.0),
            (2.0 * np.pi, +1.0, 0.0),
        ],
    )
    def test_figure_eight(self, given, expected, uncertainty):
        assert figure_eight(given) == pytest.approx(expected, uncertainty)



"""Tests for :class:`acoustic_toolbox.Octave.Octave`"""

from acoustic_toolbox.octave import Octave
import numpy as np


class TestOctave:
    """Test :class:`acoustic_toolbox.Octave.Octave`."""

    def test_interval(self):
        emin = 1.0
        emax = 4.0
        f = np.logspace(emin, emax, 50)

        o = Octave(interval=f)

        assert o.fmin == 10.0**emin
        assert o.fmax == 10.0**emax
        assert len(o.n) == len(o.center)

        o.unique = True
        assert len(o.n) == len(f)

    def test_minmax(self):
        fmin = 250.0
        fmax = 1000.0

        o = Octave(fmin=fmin, fmax=fmax)

        assert len(o.center) == 3  # 250, 500, and 1000 Hz
        assert len(o.n) == 3



import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal

import pytest

from acoustic_toolbox.room import (
    mean_alpha,
    nrc,
    t60_sabine,
    t60_eyring,
    t60_millington,
    t60_fitzroy,
    t60_arau,
    t60_impulse,
    c50_from_file,
    c80_from_file,
)
from acoustic_toolbox.bands import octave, third

from tests.get_data_path import data_path


class TestRoom:
    """Test :mod:`acoustic_toolbox.room`"""

    surfaces = np.array([240, 600, 500])
    alpha = np.array([0.1, 0.25, 0.45])
    alpha_bands = np.array(
        [[0.1, 0.1, 0.1, 0.1], [0.25, 0.25, 0.25, 0.25], [0.45, 0.45, 0.45, 0.45]]
    )
    volume = 3000

    def test_t60_sabine(self):
        calculated = t60_sabine(self.surfaces, self.alpha, self.volume)
        real = 1.211382149
        assert_almost_equal(calculated, real)

    def test_t60_sabine_bands(self):
        calculated = t60_sabine(self.surfaces, self.alpha_bands, self.volume)
        real = np.array([1.211382149, 1.211382149, 1.211382149, 1.211382149])
        assert_array_almost_equal(calculated, real)

    def test_t60_eyring(self):
        calculated = t60_eyring(self.surfaces, self.alpha, self.volume)
        real = 1.020427763
        assert_almost_equal(calculated, real)

    def test_t60_eyring_bands(self):
        calculated = t60_eyring(self.surfaces, self.alpha_bands, self.volume)
        real = np.array([1.020427763, 1.020427763, 1.020427763, 1.020427763])
        assert_array_almost_equal(calculated, real)

    def test_t60_millington(self):
        calculated = t60_millington(self.surfaces, self.alpha, self.volume)
        real = 1.020427763
        assert_almost_equal(calculated, real)

    def test_t60_millington_bands(self):
        calculated = t60_millington(self.surfaces, self.alpha_bands, self.volume)
        real = np.array([1.020427763, 1.020427763, 1.020427763, 1.020427763])
        assert_array_almost_equal(calculated, real)

    def test_t60_fitzroy(self):
        surfaces_fitzroy = np.array([240, 240, 600, 600, 500, 500])
        alpha_fitzroy = np.array([0.1, 0.1, 0.25, 0.25, 0.45, 0.45])
        calculated = t60_fitzroy(surfaces_fitzroy, alpha_fitzroy, self.volume)
        real = 0.699854185
        assert_almost_equal(calculated, real)

    def test_t60_fitzroy_bands(self):
        surfaces_fitzroy = np.array([240, 240, 600, 600, 500, 500])
        alpha_bands_f = np.array(
            [
                [0.1, 0.1, 0.25, 0.25, 0.45, 0.45],
                [0.1, 0.1, 0.25, 0.25, 0.45, 0.45],
                [0.1, 0.1, 0.25, 0.25, 0.45, 0.45],
            ]
        )
        calculated = t60_fitzroy(surfaces_fitzroy, alpha_bands_f, self.volume)
        real = np.array([0.699854185, 0.699854185, 0.699854185])
        assert_array_almost_equal(calculated, real)

    def test_t60_arau(self):
        Sx = self.surfaces[0]
        Sy = self.surfaces[1]
        Sz = self.surfaces[2]
        calculated = t60_arau(Sx, Sy, Sz, self.alpha, self.volume)
        real = 1.142442931
        assert_almost_equal(calculated, real)

    def test_t60_arau_bands(self):
        Sx = self.surfaces[0]
        Sy = self.surfaces[1]
        Sz = self.surfaces[2]
        calculated = t60_arau(Sx, Sy, Sz, self.alpha_bands, self.volume)
        real = np.array([1.142442931, 1.142442931, 1.142442931, 1.142442931])
        assert_array_almost_equal(calculated, real)

    def test_mean_alpha_float(self):
        alpha = 0.1
        surface = 10
        calculated = mean_alpha(alpha, surface)
        real = 0.1
        assert_almost_equal(calculated, real)

    def test_mean_alpha_1d(self):
        alpha = np.array([0.1, 0.2, 0.3])
        surfaces = np.array([20, 30, 40])
        calculated = mean_alpha(alpha, surfaces)
        real = 0.222222222
        assert_almost_equal(calculated, real)

    def test_mean_alpha_bands(self):
        alpha = np.array([[0.1, 0.1, 0.1], [0.2, 0.2, 0.2], [0.3, 0.3, 0.3]])
        surfaces = np.array([20, 30, 40])
        calculated = mean_alpha(alpha, surfaces)
        real = np.array([0.222222222, 0.222222222, 0.222222222])
        assert_array_almost_equal(calculated, real)

    def test_nrc_1d(self):
        alpha = np.array([0.1, 0.25, 0.5, 0.9])
        calculated = nrc(alpha)
        real = 0.4375
        assert_almost_equal(calculated, real)

    def test_nrc_2d(self):
        alphas = np.array([[0.1, 0.2, 0.3, 0.4], [0.4, 0.5, 0.6, 0.7]])
        calculated = nrc(alphas)
        real = np.array([0.25, 0.55])
        assert_array_almost_equal(calculated, real)

    @pytest.mark.parametrize(
        "file_name, bands, rt, expected",
        [
            (
                data_path() / "ir_sportscentre_omni.wav",
                octave(125, 4000),
                "t30",
                np.array([7.388, 8.472, 6.795, 6.518, 4.797, 4.089]),
            ),
            (
                data_path() / "ir_sportscentre_omni.wav",
                octave(125, 4000),
                "edt",
                np.array([4.667, 5.942, 6.007, 5.941, 5.038, 3.735]),
            ),
            (
                data_path() / "living_room_1.wav",
                octave(63, 8000),
                "t30",
                np.array([0.274, 0.365, 0.303, 0.259, 0.227, 0.211, 0.204, 0.181]),
            ),
            (
                data_path() / "living_room_1.wav",
                octave(63, 8000),
                "t20",
                np.array([0.300, 0.365, 0.151, 0.156, 0.102, 0.076, 0.146, 0.152]),
            ),
            (
                data_path() / "living_room_1.wav",
                octave(63, 8000),
                "t10",
                np.array([0.185, 0.061, 0.109, 0.024, 0.039, 0.023, 0.105, 0.071]),
            ),
            (
                data_path() / "living_room_1.wav",
                octave(63, 8000),
                "edt",
                np.array([0.267, 0.159, 0.080, 0.037, 0.021, 0.010, 0.022, 0.020]),
            ),
            (
                data_path() / "living_room_1.wav",
                third(100, 5000),
                "t30",
                np.array(
                    [
                        0.318,
                        0.340,
                        0.259,
                        0.311,
                        0.267,
                        0.376,
                        0.342,
                        0.268,
                        0.212,
                        0.246,
                        0.211,
                        0.232,
                        0.192,
                        0.231,
                        0.252,
                        0.202,
                        0.184,
                        0.216,
                    ]
                ),
            ),
            (
                data_path() / "living_room_1.wav",
                third(100, 5000),
                "t20",
                np.array(
                    [
                        0.202,
                        0.383,
                        0.189,
                        0.173,
                        0.141,
                        0.208,
                        0.323,
                        0.221,
                        0.102,
                        0.110,
                        0.081,
                        0.128,
                        0.072,
                        0.074,
                        0.087,
                        0.129,
                        0.137,
                        0.171,
                    ]
                ),
            ),
            (
                data_path() / "living_room_1.wav",
                third(100, 5000),
                "t10",
                np.array(
                    [
                        0.110,
                        0.104,
                        0.132,
                        0.166,
                        0.135,
                        0.040,
                        0.119,
                        0.223,
                        0.025,
                        0.023,
                        0.047,
                        0.050,
                        0.010,
                        0.017,
                        0.039,
                        0.084,
                        0.154,
                        0.093,
                    ]
                ),
            ),
            (
                data_path() / "living_room_1.wav",
                third(100, 5000),
                "edt",
                np.array(
                    [
                        0.354,
                        0.328,
                        0.284,
                        0.210,
                        0.132,
                        0.116,
                        0.085,
                        0.114,
                        0.064,
                        0.045,
                        0.047,
                        0.047,
                        0.024,
                        0.017,
                        0.016,
                        0.022,
                        0.020,
                        0.036,
                    ]
                ),
            ),
        ],
    )
    def test_t60_impulse(self, file_name, bands, rt, expected):
        calculated = t60_impulse(file_name, bands, rt)
        assert_array_almost_equal(calculated, expected, decimal=0)

    @pytest.mark.parametrize(
        "file_name, bands, expected",
        [
            (
                data_path() / "living_room_1.wav",
                octave(63, 8000),
                np.array([8.0, 18.0, 23.0, 26.0, 30.0, 31.0, 27.0, 29.0]),
            ),
            (
                data_path() / "living_room_1.wav",
                third(100, 5000),
                np.array(
                    [
                        3.0,
                        6.0,
                        7.0,
                        13.0,
                        18.0,
                        23.0,
                        20.0,
                        19.0,
                        28.0,
                        30.0,
                        30.0,
                        27.0,
                        32.0,
                        31.0,
                        30.0,
                        28.0,
                        29.0,
                        25.0,
                    ]
                ),
            ),
        ],
    )
    def test_c50_from_file(self, file_name, bands, expected):
        calculated = c50_from_file(file_name, bands)
        assert_array_almost_equal(calculated, expected, decimal=0)

    # TODO: Test clarity, and/or test c50/c80 with bands=None
    @pytest.mark.parametrize(
        "file_name, bands, expected",
        [
            (
                data_path() / "living_room_1.wav",
                octave(63, 8000),
                np.array(
                    [18.542, 23.077, 27.015, 31.743, 35.469, 36.836, 33.463, 36.062]
                ),
            ),
            (
                data_path() / "living_room_1.wav",
                third(100, 5000),
                np.array(
                    [
                        17.0,
                        14.0,
                        17.0,
                        24.0,
                        26.0,
                        27.0,
                        22.0,
                        26.0,
                        34.0,
                        35.0,
                        34.0,
                        32.0,
                        38.0,
                        38.0,
                        34.0,
                        34.0,
                        35.0,
                        32.0,
                    ]
                ),
            ),
        ],
    )
    def test_c80_from_file(self, file_name, bands, expected):
        calculated = c80_from_file(file_name, bands)
        assert_array_almost_equal(calculated, expected, decimal=0)



from acoustic_toolbox.decibel import dbsum, dbmean, dbadd, dbsub, dbmul, dbdiv


class TestDecibel:
    """Test :mod:`acoustic_toolbox.decibel`"""

    def test_dbsum(self):
        assert abs(dbsum([10.0, 10.0]) - 13.0103) < 1e-5

    def test_dbmean(self):
        assert dbmean([10.0, 10.0]) == 10.0

    def test_dbadd(self):
        assert abs(dbadd(10.0, 10.0) - 13.0103) < 1e-5

    def test_dbsub(self):
        assert abs(dbsub(13.0103, 10.0) - 10.0) < 1e-5

    def test_dbmul(self):
        assert abs(dbmul(10.0, 2) - 13.0103) < 1e-5

    def test_dbdiv(self):
        assert abs(dbdiv(13.0103, 2) - 10.0) < 1e-5



import pytest
import numpy as np
from acoustic_toolbox.generator import noise
from acoustic_toolbox.signal import octaves

parameters = [
    ("white", +3.0, 0.0),
    ("pink", 0.0, -3.0),
    ("blue", +6.0, +3.0),
    ("brown", -3.0, -6.0),
    ("violet", +9.0, +6.0),
]
# color, power_change, power_density_change

ERROR = 1.0
"""Permitted error for power_change and power_density_change.
This value is relatively large for `violet` because of its steep
inclination. On other cases the error is mostly < 0.1
"""


@pytest.fixture(params=parameters)
def parameters(request):
    return request.param


@pytest.fixture
def color(parameters):
    return parameters[0]


@pytest.fixture
def power_change(parameters):
    return parameters[1]


@pytest.fixture
def power_density_change(parameters):
    return parameters[2]


@pytest.fixture(params=[48000 * 10, 48000 * 10 + 1])
def samples(request):
    return request.param


class TestNoise:
    """Test :func:`acoustic_toolbox.generator.noise`."""

    def test_length(self, color, samples):
        assert len(noise(samples, color)) == samples

    def test_power(self, color, samples, power_change):
        fs = 48000
        _, L = octaves(noise(samples, color), fs)
        change = np.diff(L).mean()
        assert np.abs(change - power_change) < ERROR

    def test_power_density(self, color, samples, power_density_change):
        fs = 48000
        _, L = octaves(noise(samples, color), fs, density=True)
        change = np.diff(L).mean()
        assert np.abs(change - power_density_change) < ERROR



"""The reflection module contains functions for calculating reflection factors and impedances."""

# TODO: Need to type hint and finish docstrings
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from scipy.special import erfc  # pylint: disable=no-name-in-module

SPECIFIC_HEAT_RATIO = 1.4
r"""Specific heat ratio of air $\gamma$."""
POROSITY_DECREASE = 120.0
r"""Rate of exponential decrease of porosity with depth $\alpha$."""
SOUNDSPEED = 343.0
r"""Speed of sound in air $c$."""
DENSITY = 1.296
r"""Density of air $\rho$."""


class Boundary:
    r"""An object describing a boundary.

    Attributes:
        frequency: Frequency. Single value or vector for a frequency range.
        flow_resistivity: Flow resistivity $\sigma$.
        density: Density of air $\rho$.

            **Note:** This value is only required for when calculating the impedance according to Attenborough's model.
            See [`impedance_attenborough`][acoustic_toolbox.reflection.impedance_attenborough].

        soundspeed: Speed of sound in air $c$.

            **Note:** This value is required when calculating the impedance according to Attenborough's model or when
            calculating the spherical wave reflection factor. See respectively [`impedance_attenborough`][acoustic_toolbox.reflection.impedance_attenborough]
            and [`reflection_factor_spherical_wave`][acoustic_toolbox.reflection.reflection_factor_spherical_wave].

        porosity_decrease: Rate of exponential decrease of porosity with depth $\alpha$.

            **Note:** This value is only required for when calculating the impedance according to Attenborough's model.
            See [`impedance_attenborough`][acoustic_toolbox.reflection.impedance_attenborough].

        specific_heat_ratio: Ratio of specific heats $\gamma$ for air.

            **Note:** This value is only required for when calculating the impedance according to Attenborough's model.
            See [`impedance_attenborough`][acoustic_toolbox.reflection.impedance_attenborough].

        angle: Angle of incidence $\theta$.
        distance: Path length of the reflected ray $r$.

            **Note:** This value is only required for when calculating the spherical wave reflection factor.
            See [`reflection_factor_spherical_wave`][acoustic_toolbox.reflection.reflection_factor_spherical_wave].

        impedance_model: Impedance model.

            **Note:** Possibilities are `db` and `att` for respectively [`impedance_delany_and_bazley`][acoustic_toolbox.reflection.impedance_delany_and_bazley]
            and [`impedance_attenborough`][acoustic_toolbox.reflection.impedance_attenborough].

        reflection_model: Reflection factor model.

            **Note:** Possibilities are `plane` and `spherical` for respectively [`reflection_factor_plane_wave`][acoustic_toolbox.reflection.reflection_factor_plane_wave]
            and [`reflection_factor_spherical_wave`][acoustic_toolbox.reflection.reflection_factor_spherical_wave].
    """

    def __init__(  # pylint: disable=too-many-instance-attributes
        self,
        frequency,
        flow_resistivity,
        density=DENSITY,
        soundspeed=SOUNDSPEED,
        porosity_decrease=POROSITY_DECREASE,
        specific_heat_ratio=SPECIFIC_HEAT_RATIO,
        angle=None,
        distance=None,
        impedance_model="db",
        reflection_model="plane",
    ):
        """Initialize the boundary."""
        self.frequency = frequency
        self.flow_resistivity = flow_resistivity
        self.density = density
        self.soundspeed = soundspeed

        self.porosity_decrease = porosity_decrease

        self.specific_heat_ratio = specific_heat_ratio

        self.angle = angle
        self.distance = distance

        self.impedance_model = impedance_model
        self.reflection_model = reflection_model

    @property
    def wavenumber(self) -> float:
        r"""Wavenumber $k$.

        Returns:
            Wavenumber calculated as:
            $$
            k = \frac{2 \pi f}{c}
            $$
        """
        return 2.0 * np.pi * self.frequency / self.soundspeed

    @property
    def impedance(self) -> np.ndarray:
        """Impedance according to chosen impedance model defined using `impedance_model`.

        Raises:
            ValueError: If the impedance model is incorrect.
        """
        if self.impedance_model == "db":
            return impedance_delany_and_bazley(self.frequency, self.flow_resistivity)
        if self.impedance_model == "att":
            return impedance_attenborough(
                self.frequency,
                self.flow_resistivity,
                self.density,
                self.soundspeed,
                self.porosity_decrease,
                self.specific_heat_ratio,
            )
        else:
            raise ValueError("Incorrect impedance model.")

    @property
    def reflection_factor(self) -> np.ndarray:
        """Reflection factor according to chosen reflection factor model defined using `reflection_model`.

        Raises:
            AttributeError: If the angle is not specified.
            AttributeError: If the distance is not specified.
            ValueError: ..shrug..
        """
        if self.angle is None:
            raise AttributeError(
                "Cannot calculate reflection factor. self.angle has not been specified."
            )

        if self.reflection_model == "plane":
            return reflection_factor_plane_wave(
                *np.meshgrid(self.impedance, self.angle)
            )
        elif self.reflection_model == "spherical":
            if self.distance is None:
                raise AttributeError(
                    "Cannot calculate reflection factor. self.distance has not been specified."
                )
            else:
                return reflection_factor_spherical_wave(
                    *np.meshgrid(self.impedance, self.angle),
                    distance=self.distance,
                    wavenumber=self.wavenumber,
                )
        else:
            raise RuntimeError("Oops...")

    def plot_impedance(self, filename=None) -> Figure:
        """Plot magnitude and phase of the impedance as function of frequency.

        Args:
            filename: File name to save the plot to.

        Returns:
            Figure: The figure.
        """
        fig = plt.figure()

        ax0 = fig.add_subplot(211)
        ax0.set_title("Magnitude of impedance")
        ax0.semilogx(self.frequency, np.abs(self.impedance))
        ax0.set_xlabel(r"$f$ in Hz")
        ax0.set_ylabel(r"$\left|Z\right|$")
        ax0.grid()

        ax0 = fig.add_subplot(212)
        ax0.set_title("Angle of impedance")
        ax0.semilogx(self.frequency, np.angle(self.impedance))
        ax0.set_xlabel(r"$f$ in Hz")
        ax0.set_ylabel(r"$\angle Z$")
        ax0.grid()

        plt.tight_layout()

        if filename:
            fig.savefig(filename, transparant=True)
        return fig

    def plot_reflection_factor(self, filename=None) -> Figure:
        """Plot reflection factor.

        Args:
            filename: File name to save the plot to.

        Returns:
            Figure: The figure.
        """
        if self.frequency is None:
            raise ValueError("No frequency specified.")
        if self.angle is None:
            raise ValueError("No angle specified.")

        try:
            n_f = len(self.frequency)
        except TypeError:
            n_f = 1
        try:
            n_a = len(self.angle)
        except TypeError:
            n_a = 1

        if n_f == 1 and n_a == 1:
            raise ValueError("Either frequency or angle needs to be a vector.")

        elif n_f == 1 or n_a == 1:
            if (
                n_f == 1 and n_a > 1
            ):  # Show R as function of angle for a single frequency.
                xlabel = r"$\theta$ in degrees"
            elif (
                n_f > 1 and n_a == 1
            ):  # Show R as function of frequency for a single angle.
                xlabel = r"$f$ in Hz"
            R = self.reflection_factor
            fig = plt.figure()

            ax0 = fig.add_subplot(211)
            ax0.set_title("Magnitude of reflection factor")
            ax0.semilogx(self.frequency, np.abs(R))
            ax0.set_xlabel(xlabel)
            ax0.set_ylabel(r"$\left|R\right|$")
            ax0.grid()

            ax1 = fig.add_subplot(212)
            ax1.set_title("Phase of reflection factor")
            ax1.semilogx(self.frequency, np.angle(R))
            ax1.set_xlabel(xlabel)
            ax1.set_ylabel(r"$\angle R$")
            ax1.grid()

        elif n_f > 1 and n_a > 1:  # Show 3D or pcolor
            R = self.reflection_factor
            fig = plt.figure()

            # grid = AxesGrid(fig, 111, nrows_ncols=(2, 2), axes_pad=0.1, cbar_mode='each', cbar_location='right')
            ax0 = fig.add_subplot(211)
            # ax0 = grid[0]
            ax0.set_title("Magnitude of reflection factor")
            ax0.pcolormesh(self.frequency, self.angle * 180.0 / np.pi, np.abs(R))
            # ax0.pcolor(self.angle, self.frequency, np.abs(R))
            # ax0.set_xlabel(xlabel)
            # ax0.set_ylabel(r'$\left|R\right|$')
            ax0.grid()

            ax1 = fig.add_subplot(212)
            # ax1 = grid[1]
            ax1.set_title("Phase of reflection factor")
            ax1.pcolormesh(self.frequency, self.angle * 180.0 / np.pi, np.angle(R))
            # ax1.pcolor(self.angle, self.frequency, np.angle(R))
            # ax0.set_xlabel(xlabel)
            # ax0.set_ylabel(r'$\angle R$')
            ax1.grid()

        else:
            raise RuntimeError("Oops...")

        # plt.tight_layout()

        if filename:
            fig.savefig(filename, transparant=True)
        else:
            return fig


def reflection_factor_plane_wave(impedance, angle):
    r"""Plane wave reflection factor $R$.

    Args:
      impedance: Normalized impedance $Z$.
      angle: Angle of incidence $\theta$.

    Returns:
        float: Plane wave reflection factor calculated as:
            $$
            R = \frac{Z \cos{\theta} - 1}{Z \cos{\theta} + 1}
            $$
    """
    return (impedance * np.cos(angle) - 1.0) / (impedance * np.cos(angle) + 1.0)


def numerical_distance(impedance, angle, distance, wavenumber):
    r"""Numerical distance $w$.

    Args:
      impedance: Normalized impedance $Z$.
      angle: Angle of incidence $\theta$.
      distance: Path length of the reflected ray $r$.
      wavenumber: Wavenumber $k$.

    Returns:
        complex: Numerical distance calculated as:
            $$
            w = \sqrt{-j k r  \left( 1 + \frac{1}{Z} \cos{\theta} - \sqrt{1 - \left( \frac{1}{Z} \right)^2} \sin{\theta} \right) }
            $$
    """
    return np.sqrt(
        -1j
        * wavenumber
        * distance
        * (
            1.0
            + 1.0 / impedance * np.cos(angle)
            - np.sqrt(1.0 - (1.0 / impedance) ** 2.0) * np.sin(angle)
        )
    )


def reflection_factor_spherical_wave(impedance, angle, distance, wavenumber):
    r"""Spherical wave reflection factor $Q$.

    Args:
      impedance: Normalized impedance $Z$.
      angle: Angle of incidence $\theta$.
      distance: Path length of the reflected ray $r$.
      wavenumber: Wavenumber $k$.

    Returns:
        complex: Spherical wave reflection factor calculated as:
            $$
            Q = R \left(1 - R \right) F
            $$

            where $R$ is the plane wave reflection factor as calculated in [reflection_factor_plane_wave][acoustic_toolbox.reflection.reflection_factor_plane_wave] and $F$ is given by
            $$
            F = 1 - j \sqrt{\pi} w e^{-w^2} \mathrm{erfc} \left( j w \right)
            $$
    """
    w = numerical_distance(impedance, angle, distance, wavenumber)

    F = 1.0 - 1j * np.sqrt(np.pi) * w * np.exp(-(w**2.0)) * erfc(1j * w)

    plane_factor = reflection_factor_plane_wave(impedance, angle)
    return plane_factor * (1.0 - plane_factor) * F


def impedance_delany_and_bazley(frequency, flow_resistivity):
    r"""Normalised specific acoustic impedance according to the empirical one-parameter model by Delany and Bazley.

    Args:
      frequency: Frequency $f$.
      flow_resistivity: Flow resistivity $\sigma$.

    Returns:
        complex: Impedance calculated as:
            $$
            Z = 1 + 9.08 \left( \frac{1000f}{\sigma} \right)^{-0.75} - 11.9 j \left( \frac{1000f}{\sigma} \right)^{-0.73}
            $$
    """
    return (
        1.0
        + 9.08 * (1000.0 * frequency / flow_resistivity) ** (-0.75)
        - 1j * 11.9 * (1000.0 * frequency / flow_resistivity) ** (-0.73)
    )


def impedance_attenborough(
    frequency,
    flow_resistivity,
    density=DENSITY,
    soundspeed=SOUNDSPEED,
    porosity_decrease=POROSITY_DECREASE,
    specific_heat_ratio=SPECIFIC_HEAT_RATIO,
):
    r"""Normalised specific acoustics impedance according to the two-parameter model by Attenborough.

    Args:
      frequency: Frequency $f$.
      flow_resistivity: Flow resistivity $\sigma$.
      soundspeed: Speed of sound in air $c$.
      density: Density of air $\rho$.
      porosity_decrease: Rate of exponential decrease of porosity with depth $\alpha$.
      specific_heat_ratio: Ratio of specific heats $\gamma$ for air.

    Returns:
        complex: Impedance calculated as:
            $$
            Z = \frac{\left( 1-j \right) \sqrt{\sigma/f}}{\sqrt{\pi \gamma_0 \rho_0}} - \frac{jc\alpha}{8 \pi \gamma_0 f}
            $$
    """
    return (1.0 - 1j) * np.sqrt(flow_resistivity / frequency) / np.sqrt(
        np.pi * specific_heat_ratio * density
    ) - 1j * soundspeed * porosity_decrease / (
        8.0 * np.pi * specific_heat_ratio * frequency
    )



"""Utility functions."""

import numpy as np
from acoustic_toolbox.decibel import dbsum

SOUNDSPEED = 343.0
"""Speed of sound in air."""

esum = dbsum


def mean_tl(tl, surfaces):
    """Mean tl."""
    try:
        tau_axis = tl.ndim - 1
    except AttributeError:
        tau_axis = 0
    tau = 1.0 / (10.0 ** (tl / 10.0))
    return 10.0 * np.log10(1.0 / np.average(tau, tau_axis, surfaces))


def wavelength(freq: np.ndarray, c: float = SOUNDSPEED) -> np.ndarray:
    """Wavelength for one or more frequencies."""
    return c / freq


def w(freq: np.ndarray) -> np.ndarray:
    """Angular frequency for one o more frequencies."""
    return 2.0 * np.pi * freq


def _is_1d(given: np.ndarray):
    """Check if the given object is a 1D array."""
    if isinstance(given, (int, float)):
        return given
    elif given.ndim == 1:
        return np.array([given])
    elif given.ndim == 2 and given.shape[0] == 1:
        return given[0]
    else:
        return given



"""# ISO 9613-1:1993

This module implements ISO 9613-1:1993 which provides analytical methods for calculating
sound attenuation due to atmospheric absorption under various meteorological conditions.

The module provides functions for calculating:
- Sound speed in different temperatures
- Saturation vapor pressure
- Molar concentration of water vapor
- Relaxation frequencies of oxygen and nitrogen
- Atmospheric absorption coefficients

Constants:
    [SOUNDSPEED][acoustic_toolbox.standards.iso_9613_1_1993.SOUNDSPEED]: Speed of sound constant.

    [REFERENCE_TEMPERATURE][acoustic_toolbox.standards.iso_9613_1_1993.REFERENCE_TEMPERATURE]: Reference temperature constant.

    [REFERENCE_PRESSURE][acoustic_toolbox.standards.iso_9613_1_1993.REFERENCE_PRESSURE]: Reference pressure constant.

    [TRIPLE_TEMPERATURE][acoustic_toolbox.standards.iso_9613_1_1993.TRIPLE_TEMPERATURE]: Triple point isotherm temperature constant.

Reference:
    ISO 9613-1:1993: Acoustics — Attenuation of sound during propagation outdoors
"""

import numpy as np
from numpy.typing import NDArray

SOUNDSPEED = 343.2
"""Speed of sound."""

REFERENCE_TEMPERATURE = 293.15
"""Reference temperature."""

REFERENCE_PRESSURE = 101.325
"""International Standard Atmosphere pressure in kilopascal."""

TRIPLE_TEMPERATURE = 273.16
"""Triple point isotherm temperature."""


def soundspeed(
    temperature: float, reference_temperature: float = REFERENCE_TEMPERATURE
) -> float:
    r"""Calculate speed of sound $c$ in air.

    Notes:
        The speed of sound is calculated using the formula:
        $$
        c = 343.2 \left( \frac{T}{T_0} \right)
        $$

    Args:
        temperature: Ambient temperature $T_0$.
        reference_temperature: Reference temperature $T$.

    Returns:
        float: Speed of sound in m/s.
    """
    return 343.2 * np.sqrt(temperature / reference_temperature)


def saturation_pressure(
    temperature: float,
    reference_pressure: float = REFERENCE_PRESSURE,
    triple_temperature: float = TRIPLE_TEMPERATURE,
) -> float:
    r"""Calculate saturation vapor pressure $p_{sat}$.

    Notes:
        The saturation vapor pressure is calculated using the formula:
        $$
        p_{sat} = 10^C \cdot p_r
        $$

        with exponent $C$ given by:
        $$
        C = -6.8346 \cdot \left( \frac{T_{01}}{T} \right)^{1.261}  + 4.6151
        $$

    Args:
        temperature: Ambient temperature $T$.
        reference_pressure: Reference pressure $p_r$.
        triple_temperature: Triple point temperature $T_{01}$.

    Returns:
        float: Saturation vapor pressure
    """
    return reference_pressure * 10.0 ** (
        -6.8346 * (triple_temperature / temperature) ** (1.261) + 4.6151
    )


def molar_concentration_water_vapour(
    relative_humidity: float, saturation_pressure: float, pressure: float
) -> float:
    r"""Calculate molar concentration of water vapour $h$.

    Notes:
        The molar concentration of water vapour is calculated using the formula:
        $$
        h = h_r  \frac{p_{sat}}{p_a}
        $$

    Args:
        relative_humidity: Relative humidity $h_r$
        saturation_pressure: Saturation pressure $p_{sat}$
        pressure: Ambient pressure $p_a$

    Returns:
        float: Molar concentration of water vapour
    """
    return relative_humidity * saturation_pressure / pressure


def relaxation_frequency_oxygen(
    pressure: float, h: float, reference_pressure: float = REFERENCE_PRESSURE
) -> float:
    r"""Calculate relaxation frequency of oxygen $f_{r,O}$.

    Notes:
        The relaxation frequency of oxygen is calculated using the formula:
        $$
        f_{r,O} = \frac{p_a}{p_r} \left( 24 + 4.04 \cdot 10^4 h \frac{0.02 + h}{0.391 + h}  \right)
        $$

    Args:
        pressure: Ambient pressure $p_a$
        h: Molar concentration of water vapour $h$
        reference_pressure: Reference pressure $p_r$.

    Returns:
        float: Relaxation frequency of oxygen in Hz

    See Also:
        [molar_concentration_water_vapour][acoustic_toolbox.standards.iso_9613_1_1993.molar_concentration_water_vapour]: Function to calculate molar concentration of water vapour
    """
    return (
        pressure
        / reference_pressure
        * (24.0 + 4.04 * 10.0**4.0 * h * (0.02 + h) / (0.391 + h))
    )


def relaxation_frequency_nitrogen(
    pressure: float,
    temperature: float,
    h: float,
    reference_pressure: float = REFERENCE_PRESSURE,
    reference_temperature: float = REFERENCE_TEMPERATURE,
) -> float:
    r"""Calculate relaxation frequency of nitrogen $f_{r,N}$.

    Notes:
        The relaxation frequency of nitrogen is calculated by:
        $$
        f_{r,N} = \frac{p_a}{p_r} \left( \frac{T}{T_0} \right)^{-1/2} \cdot \left( 9 + 280 \cdot h \cdot \exp \left( -4.170 \left[ \left(\frac{T}{T_0} \right)^{-1/3} -1 \right] \right) \right)
        $$

    Args:
        pressure: Ambient pressure $p_a$
        temperature: Ambient temperature $T$
        h: Molar concentration of water vapour $h$
        reference_pressure: Reference pressure $p_{ref}$.
        reference_temperature: Reference temperature $T_{ref}$.

    Returns:
        float: Relaxation frequency of nitrogen in Hz
    """
    return (
        pressure
        / reference_pressure
        * (temperature / reference_temperature) ** (-0.5)
        * (
            9.0
            + 280.0
            * h
            * np.exp(
                -4.170 * ((temperature / reference_temperature) ** (-1.0 / 3.0) - 1.0)
            )
        )
    )


def attenuation_coefficient(
    pressure: float,
    temperature: float,
    reference_pressure: float,
    reference_temperature: float,
    relaxation_frequency_nitrogen: float,
    relaxation_frequency_oxygen: float,
    frequency: float | NDArray[np.float64],
) -> float | NDArray[np.float64]:
    r"""Calculate attenuation coefficient $\alpha$ describing atmospheric absorption in dB/m for the specified ``frequency``.

    Args:
        pressure: Ambient pressure $p_a$
        temperature: Ambient temperature $T$
        reference_pressure: Reference pressure $p_{ref}$
        reference_temperature: Reference temperature $T_{ref}$
        relaxation_frequency_nitrogen: Relaxation frequency of nitrogen $f_{r,N}$
        relaxation_frequency_oxygen: Relaxation frequency of oxygen $f_{r,O}$
        frequency: Frequencies $f$ to calculate $\alpha$ for

    Returns:
        float | NDArray[np.float64]: Attenuation coefficient in dB/m
    """
    return (
        8.686
        * frequency**2.0
        * (
            (
                1.84
                * 10.0 ** (-11.0)
                * (reference_pressure / pressure)
                * (temperature / reference_temperature) ** (0.5)
            )
            + (temperature / reference_temperature) ** (-2.5)
            * (
                0.01275
                * np.exp(-2239.1 / temperature)
                * (
                    relaxation_frequency_oxygen
                    + (frequency**2.0 / relaxation_frequency_oxygen)
                )
                ** (-1.0)
                + 0.1068
                * np.exp(-3352.0 / temperature)
                * (
                    relaxation_frequency_nitrogen
                    + (frequency**2.0 / relaxation_frequency_nitrogen)
                )
                ** (-1.0)
            )
        )
    )



"""Module for working with octaves.

The following is an example on how to use [acoustic_toolbox.octave.Octave][acoustic_toolbox.octave.Octave].

.. literalinclude:: ../../examples/example_octave.py

"""

import numpy as np
from acoustic_toolbox.standards import iec_61260_1_2014
from acoustic_toolbox.standards.iec_61260_1_2014 import index_of_frequency
from acoustic_toolbox.standards.iec_61260_1_2014 import REFERENCE_FREQUENCY as REFERENCE

# REFERENCE = 1000.0
# """
# Reference frequency.
# """

# from acoustic_toolbox.standards.iec_61260_1_2014 import index_of_frequency
# from acoustic_toolbox.standards.iec_61260_1_2014 import REFERENCE_FREQUENCY as REFERENCE


def exact_center_frequency(frequency=None, fraction=1, n=None, ref=REFERENCE):
    """Exact center frequency.

    Args:
      frequency: Frequency within the band.
      fraction: Band designator.
      n: Index of band.
      ref: Reference frequency.

    Returns:
      Exact center frequency for the given frequency or band index.

    See Also:
      - [acoustic_toolbox.standards.iec_61260_1_2014.exact_center_frequency][acoustic_toolbox.standards.iec_61260_1_2014.exact_center_frequency]
      - [acoustic_toolbox.standards.iec_61260_1_2014.index_of_frequency][acoustic_toolbox.standards.iec_61260_1_2014.index_of_frequency]
    """
    if frequency is not None:
        n = iec_61260_1_2014.index_of_frequency(frequency, fraction=fraction, ref=ref)
    return iec_61260_1_2014.exact_center_frequency(n, fraction=fraction, ref=ref)


def nominal_center_frequency(frequency=None, fraction=1, n=None):
    """Nominal center frequency.

    Note:
      Contrary to the other functions this function silently assumes 1000 Hz reference frequency.

    Args:
      frequency: Frequency within the band.
      fraction: Band designator.
      n: Index of band.

    Returns:
      The nominal center frequency for the given frequency or band index.

    See Also:
      - [acoustic_toolbox.standards.iec_61260_1_2014.exact_center_frequency][acoustic_toolbox.standards.iec_61260_1_2014.exact_center_frequency]
      - [acoustic_toolbox.standards.iec_61260_1_2014.nominal_center_frequency][acoustic_toolbox.standards.iec_61260_1_2014.nominal_center_frequency]


    """
    center = exact_center_frequency(frequency, fraction, n)
    return iec_61260_1_2014.nominal_center_frequency(center, fraction)


def lower_frequency(
    frequency=None, fraction=1, n=None, ref=REFERENCE
) -> float | np.ndarray:
    """Lower band-edge frequency.

    Args:
      frequency: Frequency within the band.
      fraction: Band designator.
      n: Index of band.
      ref: Reference frequency.

    Returns:
      Lower band-edge frequency for the given frequency or band index.

    See Also:
      - [acoustic_toolbox.standards.iec_61260_1_2014.exact_center_frequency][acoustic_toolbox.standards.iec_61260_1_2014.exact_center_frequency]
      - [acoustic_toolbox.standards.iec_61260_1_2014.lower_frequency][acoustic_toolbox.standards.iec_61260_1_2014.lower_frequency]
    """
    center = exact_center_frequency(frequency, fraction, n, ref=ref)
    return iec_61260_1_2014.lower_frequency(center, fraction)


def upper_frequency(
    frequency=None, fraction=1, n=None, ref=REFERENCE
) -> float | np.ndarray:
    """Upper band-edge frequency.

    Args:
      frequency: Frequency within the band.
      fraction: Band designator.
      n: Index of band.
      ref: Reference frequency.

    Returns:
      Upper band-edge frequency for the given frequency or band index.

    See Also:
      - [acoustic_toolbox.standards.iec_61260_1_2014.exact_center_frequency][acoustic_toolbox.standards.iec_61260_1_2014.exact_center_frequency]
      - [acoustic_toolbox.standards.iec_61260_1_2014.upper_frequency][acoustic_toolbox.standards.iec_61260_1_2014.upper_frequency]
    """
    center = exact_center_frequency(frequency, fraction, n, ref=ref)
    return iec_61260_1_2014.upper_frequency(center, fraction)


# -- things below will be deprecated?---#

frequency_of_band = iec_61260_1_2014.exact_center_frequency
"""Calculate the center frequency for a given band index."""

band_of_frequency = iec_61260_1_2014.index_of_frequency
"""Calculate the band index for a given frequency."""


class Octave:
    """Class to calculate octave center frequencies.

    Attributes:
      reference: Reference center frequency $f_{c,0}$.
      fraction: Fraction of octave.
      interval: Interval.
      fmin: Minimum frequency of a range.
      fmax: Maximum frequency of a range.
      unique: Whether or not to calculate the requested values for every value of ``interval``.
      reference: Reference frequency.
    """

    def __init__(
        self,
        fraction=1,
        interval=None,
        fmin=None,
        fmax=None,
        unique: bool = False,
        reference=REFERENCE,
    ):
        """Initialize the Octave class.

        Raises:
          AttributeError: If ``interval`` is not ``None`` and ``fmin`` or ``fmax`` is not ``None``.
        """
        self.reference = reference
        self.fraction = fraction

        if (interval is not None) and (fmin is not None or fmax is not None):
            raise AttributeError("Cannot specify both interval and fmin/fmax")
        self._interval = np.asarray(interval)
        self._fmin = fmin
        self._fmax = fmax
        self.unique = unique

    @property
    def fmin(self):
        """Minimum frequency of an interval."""
        if self._fmin is not None:
            return self._fmin
        elif self._interval is not None:
            return self.interval.min()
        else:
            raise ValueError("Incorrect fmin/interval")

    @fmin.setter
    def fmin(self, x):
        if self.interval is not None:
            pass  # Warning, remove interval first.
        else:
            self._fmin = x

    @property
    def fmax(self):
        """Maximum frequency of an interval."""
        if self._fmax is not None:
            return self._fmax
        elif self._interval is not None:
            return self.interval.max()
        else:
            raise ValueError("Incorrect fmax/interval")

    @fmax.setter
    def fmax(self, x):
        if self.interval is not None:
            pass
        else:
            self._fmax = x

    @property
    def interval(self):
        """Interval."""
        return self._interval

    @interval.setter
    def interval(self, x):
        if self._fmin or self._fmax:
            pass
        else:
            self._interval = np.asarray(x)

    def _n(self, f):
        """Calculate the band ``n`` from a given frequency.

        Args:
          f: Frequency

        See also [band_of_frequency][acoustic_toolbox.octave.band_of_frequency].
        """
        return band_of_frequency(f, fraction=self.fraction, ref=self.reference)

    def _fc(self, n):
        """Calculate center frequency of band ``n``.

        Args:
          n: band ``n`.

        See also [frequency_of_band][acoustic_toolbox.octave.frequency_of_band].
        """
        return frequency_of_band(n, fraction=self.fraction, ref=self.reference)

    @property
    def n(self):
        """Return band ``n`` for a given frequency."""
        if self.interval is not None and self.unique:
            return self._n(self.interval)
        else:
            return np.arange(self._n(self.fmin), self._n(self.fmax) + 1)

    @property
    def center(self) -> float:
        r"""Return center frequency $f_c$.

        Returns:
          Center frequency calculated as:
            $$
            f_c = f_{ref} \cdot 2^{n/N} \cdot 10^{\frac{3}{10N}}
            $$
        """
        n = self.n
        return self._fc(n)

    @property
    def bandwidth(self):
        """Bandwidth of bands.

        $$
        B = f_u - f_l
        $$
        """
        return self.upper - self.lower

    @property
    def lower(self) -> float:
        r"""Lower frequency limits of bands.

        Returns:
          Lower frequency calculated as:
            $$
            f_l = f_c \cdot 2^{\frac{-1}{2N}}
            $$

        See also [lower_frequency][acoustic_toolbox.octave.lower_frequency].
        """
        return lower_frequency(self.center, self.fraction)

    @property
    def upper(self):
        r"""Upper frequency limits of bands.

        Returns:
          float: Upper frequency calculated as:
            $$
            f_u = f_c \cdot 2^{\frac{1}{2N}}
            $$

        See Also:
          [upper_frequency][acoustic_toolbox.octave.upper_frequency].
        """
        return upper_frequency(self.center, self.fraction)


__all__ = [
    "exact_center_frequency",
    "nominal_center_frequency",
    "lower_frequency",
    "upper_frequency",
    "index_of_frequency",
    "Octave",
    "frequency_of_band",
    "band_of_frequency",  # These three will be deprecated?
]



"""Weightings according to [IEC 61672-1:2003](https://webstore.iec.ch/en/publication/19903).

Attributes:
    THIRD_OCTAVE_A_WEIGHTING: A-weighting filter for preferred 1/3-octave band center frequencies.
    THIRD_OCTAVE_C_WEIGHTING: C-weighting filter for preferred 1/3-octave band center frequencies.

Functions:
    a_weighting: A-weighting.
    c_weighting: C-weighting.
    z2a: Apply A-weighting to Z-weighted signal.
    a2z: Remove A-weighting from A-weighted signal.
    z2c: Apply C-weighting to Z-weighted signal.
    c2z: Remove C-weighting from C-weighted signal.
"""

import numpy as np

from acoustic_toolbox.bands import third

THIRD_OCTAVE_A_WEIGHTING = np.array(
    [
        -63.4,
        -56.7,
        -50.5,
        -44.7,
        -39.4,
        -34.6,
        -30.2,
        -26.2,
        -22.5,
        -19.1,
        -16.1,
        -13.4,
        -10.9,
        -8.6,
        -6.6,
        -4.8,
        -3.2,
        -1.9,
        -0.8,
        +0.0,
        +0.6,
        +1.0,
        +1.2,
        +1.3,
        +1.2,
        +1.0,
        +0.5,
        -0.1,
        -1.1,
        -2.5,
        -4.3,
        -6.6,
        -9.3,
    ]
)
"""A-weighting filter for preferred 1/3-octave band center frequencies, as specified in [`acoustic_toolbox.bands.THIRD_OCTAVE_CENTER_FREQUENCIES`][acoustic_toolbox.bands.THIRD_OCTAVE_CENTER_FREQUENCIES]."""

THIRD_OCTAVE_C_WEIGHTING = np.array(
    [
        -11.2,
        -8.5,
        -6.2,
        -4.4,
        -3.0,
        -2.0,
        -1.3,
        -0.8,
        -0.5,
        -0.3,
        -0.2,
        -0.1,
        +0.0,
        +0.0,
        +0.0,
        +0.0,
        +0.0,
        +0.0,
        +0.0,
        +0.0,
        +0.0,
        -0.1,
        -0.2,
        -0.3,
        -0.5,
        -0.8,
        -1.3,
        -2.0,
        -3.0,
        -4.4,
        -6.2,
        -8.5,
        -11.2,
    ]
)
"""C-weighting filter for preferred 1/3-octave band center frequencies, as specified in [`acoustic_toolbox.bands.THIRD_OCTAVE_CENTER_FREQUENCIES`][acoustic_toolbox.bands.THIRD_OCTAVE_CENTER_FREQUENCIES]."""


def a_weighting(first: float, last: float) -> np.ndarray:
    """Select frequency weightings between ``first`` and ``last`` centerfrequencies from A-weighting.

    Possible values for these frequencies are third-octave frequencies
    between 12.5 Hz and 20,000 Hz (including them).

    Args:
      first: First third-octave centerfrequency.
      last: Last third-octave centerfrequency.

    Returns:
        A-weighting between `first` and `last` center frequencies.
    """
    return _weighting("a", first, last)


def c_weighting(first: float, last: float) -> np.ndarray:
    """Select frequency weightings between ``first`` and ``last`` centerfrequencies from C-weighting.

    Possible values for these frequencies are third-octave frequencies
    between 12.5 Hz and 20,000 Hz (including them).

    Args:
      first: First third-octave centerfrequency.
      last: Last third-octave centerfrequency.

    Returns:
        C-weighting between `first` and `last` center frequencies.
    """
    return _weighting("c", first, last)


def _weighting(filter_type: str, first: float, last: float) -> np.ndarray:
    third_oct_bands = third(12.5, 20000.0).tolist()
    low = third_oct_bands.index(first)
    high = third_oct_bands.index(last)

    if filter_type == "a":
        freq_weightings = THIRD_OCTAVE_A_WEIGHTING

    elif filter_type == "c":
        freq_weightings = THIRD_OCTAVE_C_WEIGHTING

    return freq_weightings[low : high + 1]


def z2a(levels: np.ndarray, first: float, last: float) -> np.ndarray:
    """Apply A-weighting to Z-weighted signal.

    Args:
      levels: Z-weighted signal.
      first: First third-octave centerfrequency.
      last: Last third-octave centerfrequency.

    Returns:
        A-weighted signal.
    """
    return levels + a_weighting(first, last)


def a2z(levels: np.ndarray, first: float, last: float) -> np.ndarray:
    """Remove A-weighting from A-weighted signal.

    Args:
      levels: A-weighted signal.
      first: First third-octave centerfrequency.
      last: Last third-octave centerfrequency.

    Returns:
        Z-weighted signal.
    """
    return levels - a_weighting(first, last)


def z2c(levels: np.ndarray, first: float, last: float) -> np.ndarray:
    """Apply C-weighting to Z-weighted signal.

    Args:
      levels: Z-weighted signal.
      first: First third-octave centerfrequency.
      last: Last third-octave centerfrequency.

    Returns:
        C-weighted signal.
    """
    return levels + c_weighting(first, last)


def c2z(levels: np.ndarray, first: float, last: float) -> np.ndarray:
    """Remove C-weighting from C-weighted signal.

    Args:
      levels: C-weighted signal.
      first: First third-octave centerfrequency.
      last: Last third-octave centerfrequency.

    Returns:
        Z-weighted signal.
    """
    return levels - c_weighting(first, last)


def a2c(levels: np.ndarray, first: float, last: float) -> np.ndarray:
    """Go from A-weighted to C-weighted.

    Args:
      levels: A-weighted signal.
      first: First third-octave centerfrequency.
      last: Last third-octave centerfrequency.

    Returns:
        C-weighted signal.
    """
    dB = a2z(levels, first, last)
    return z2c(dB, first, last)


def c2a(levels: np.ndarray, first: float, last: float) -> np.ndarray:
    """Go from C-weighted to A-weighted.

    Args:
      levels: C-weighted signal.
      first: First third-octave centerfrequency.
      last: Last third-octave centerfrequency.

    Returns:
        A-weighted signal.
    """
    dB = c2z(levels, first, last)
    return z2a(dB, first, last)


__all__ = [
    "THIRD_OCTAVE_A_WEIGHTING",
    "THIRD_OCTAVE_C_WEIGHTING",
    "a_weighting",
    "c_weighting",
    "z2a",
    "a2z",
    "z2c",
    "c2z",
    "a2c",
    "c2a",
]



"""The signal module contains all kinds of signal processing related functions.

# Filtering

Classes:
    Filterbank: Fractional-Octave filter bank.

Functions:
    bandpass_filter: Band-pass filter.
    bandpass: Filter signal with band-pass filter.
    lowpass: Filter signal with low-pass filter.
    highpass: Filter signal with high-pass filter.
    octave_filter: Fractional-octave band-pass filter.
    convolve: Perform convolution of a signal with a linear time-variant system.

# Windowing

Functions:
    window_scaling_factor: Calculate window scaling factor.
    apply_window: Apply window to signal.

# Spectra

Functions:
    amplitude_spectrum: Amplitude spectrum of instantaneous signal.
    auto_spectrum: Auto-spectrum of instantaneous signal.
    power_spectrum: Power spectrum of instantaneous signal.
    density_spectrum: Density spectrum of instantaneous signal.
    angle_spectrum: Phase angle spectrum of instantaneous signal.
    phase_spectrum: Phase spectrum of instantaneous signal.

# Frequency bands

Classes:
    Frequencies: Object describing frequency bands.
    EqualBand: Equal bandwidth spectrum.
    OctaveBand: Fractional-octave band spectrum.

Functions:
    integrate_bands: Reduce frequency resolution of power spectrum.
    octaves: Calculate level per 1/1-octave in frequency domain.
    third_octaves: Calculate level per 1/3-octave in frequency domain.

# Hilbert transform

Functions:
    amplitude_envelope: Instantaneous amplitude of tone.
    instantaneous_phase: Instantaneous phase of tone.
    instantaneous_frequency: Determine instantaneous frequency of tone.

# Conversion

Functions:
    decibel_to_neper: Convert decibel to neper.
    neper_to_decibel: Convert neper to decibel.

# Other

Functions:
    isolate: Isolate signals.
    zero_crossings: Determine the positions of zero crossings in data.
    rms: Root mean squared of signal.
    ms: Mean value of signal squared.
    normalize: Normalize power in signal.
    ir2fr: Convert impulse response into frequency response.
    wvd: Wigner-Ville Distribution.
"""

from __future__ import annotations
import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray
from typing import Generator
from scipy.sparse import spdiags
from scipy.signal import (
    butter,
    lfilter,
    freqz,
    filtfilt,
    sosfilt,
    lti,
    cheby1,
    firwin,
    hilbert,
)
from scipy.signal._arraytools import even_ext, odd_ext, const_ext

import acoustic_toolbox.octave
import acoustic_toolbox.bands
from acoustic_toolbox.standards.iso_tr_25417_2007 import REFERENCE_PRESSURE
from acoustic_toolbox.standards.iec_61672_1_2013 import (
    NOMINAL_OCTAVE_CENTER_FREQUENCIES,
    NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES,
)

try:
    from pyfftw.interfaces.numpy_fft import rfft  # type: ignore
except ImportError:
    from numpy.fft import rfft


def bandpass_filter(
    lowcut: float, highcut: float, fs: float, order: int = 8, output: str = "sos"
) -> tuple | None:
    """Band-pass filter.

    Args:
        lowcut: Lower cut-off frequency.
        highcut: Upper cut-off frequency.
        fs: Sample frequency.
        order: Filter order. Defaults to 8.
        output: Output type. {'ba', 'zpk', 'sos'}.

    Returns:
        tuple: Filter coefficients depending on `output`.

    See Also:
        [`scipy.signal.butter`][scipy.signal.butter]: For more details on the Butterworth filter.
    """
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    return butter(order / 2, [low, high], btype="band", output=output)


def bandpass(
    signal: np.ndarray,
    lowcut: float,
    highcut: float,
    fs: float,
    order: int = 8,
    zero_phase: bool = False,
) -> np.ndarray:
    """Filter signal with band-pass filter.

    Args:
        signal: Signal to be filtered.
        lowcut: Lower cut-off frequency.
        highcut: Upper cut-off frequency.
        fs: Sample frequency.
        order: Filter order. Defaults to 8.
        zero_phase: If True, uses `filtfilt` to prevent phase error.

    Returns:
        Filtered signal.

    See Also:
        [`bandpass_filter`][acoustic_toolbox.signal.bandpass_filter]: The filter that is used.
    """
    sos = bandpass_filter(lowcut, highcut, fs, order, output="sos")
    if zero_phase:
        return _sosfiltfilt(sos, signal)
    else:
        return sosfilt(sos, signal)


def bandstop(signal, lowcut, highcut, fs, order=8, zero_phase: bool = False):
    """Filter signal with band-stop filter.

    Args:
        signal: Signal to be filtered.
        lowcut: Lower cut-off frequency.
        highcut: Upper cut-off frequency.
        fs: Sample frequency.
        order: Filter order. Defaults to 8.
        zero_phase: If True, uses filtfilt to prevent phase error. Defaults to False.

    Returns:
        Filtered signal.

    See Also:
        [`lowpass`][acoustic_toolbox.signal.lowpass], [`highpass`][acoustic_toolbox.signal.highpass]: Used to create the band-stop filter.
    """
    return lowpass(
        signal, lowcut, fs, order=(order // 2), zero_phase=zero_phase
    ) + highpass(signal, highcut, fs, order=(order // 2), zero_phase=zero_phase)


def lowpass(signal, cutoff, fs, order=4, zero_phase: bool = False):
    """Filter signal with low-pass filter.

    A Butterworth filter is used. Filtering is done with second-order sections.

    Args:
        signal: Signal to be filtered.
        cutoff: Cut-off frequency.
        fs: Sample frequency.
        order: Filter order. Defaults to 4.
        zero_phase: If True, uses filtfilt to prevent phase error. Defaults to False.

    Returns:
        Filtered signal.

    See Also:
        [`scipy.signal.butter`][scipy.signal.butter]: For more details on the Butterworth filter.
    """
    sos = butter(order, cutoff / (fs / 2.0), btype="low", output="sos")
    if zero_phase:
        return _sosfiltfilt(sos, signal)
    else:
        return sosfilt(sos, signal)


def highpass(signal, cutoff, fs, order=4, zero_phase: bool = False):
    """Filter signal with high-pass filter.

    A Butterworth filter is used. Filtering is done with second-order sections.

    Args:
        signal: Signal to be filtered.
        cutoff: Cut-off frequency.
        fs: Sample frequency.
        order: Filter order. Defaults to 4.
        zero_phase: If True, uses filtfilt to prevent phase error. Defaults to False.

    Returns:
        Filtered signal.

    See Also:
        [`scipy.signal.butter`][scipy.signal.butter]: For more details on the Butterworth filter.
    """
    sos = butter(order, cutoff / (fs / 2.0), btype="high", output="sos")
    if zero_phase:
        return _sosfiltfilt(sos, signal)
    else:
        return sosfilt(sos, signal)


def octave_filter(center, fs, fraction, order=8, output: str = "sos"):
    """Fractional-octave band-pass filter.

    A Butterworth filter is used.
    Args:
        center: Center frequency of fractional-octave band.
        fs: Sample frequency.
        fraction: Fraction of fractional-octave band.
        order: Filter order. Defaults to 8.
        output: Output type. {'ba', 'zpk', 'sos'}. Defaults to 'sos'.

    Returns:
        tuple: Filter coefficients depending on `output`.

    See Also:
        [`bandpass_filter`][acoustic_toolbox.signal.bandpass_filter]: Used to create the fractional-octave filter.
    """
    ob = OctaveBand(center=center, fraction=fraction)
    return bandpass_filter(
        ob._get_scalar(ob.lower), ob._get_scalar(ob.upper), fs, order, output=output
    )


def octavepass(signal, center, fs, fraction, order=8, zero_phase: bool = True):
    """Filter signal with fractional-octave bandpass filter.

    A Butterworth filter is used. Filtering is done with second-order sections.

    Args:
        signal: Signal to be filtered.
        center: Center frequency of fractional-octave band.
        fs: Sample frequency.
        fraction: Fraction of fractional-octave band.
        order: Filter order. Defaults to 8.
        zero_phase: If True, uses filtfilt to prevent phase error. Defaults to True.

    Returns:
        Filtered signal.

    See Also:
        [`octave_filter`][acoustic_toolbox.signal.octave_filter]: The filter that is used.
    """
    sos = octave_filter(center, fs, fraction, order)
    if zero_phase:
        return _sosfiltfilt(sos, signal)
    else:
        return sosfilt(sos, signal)


def convolve(signal, ltv: np.ndarray, mode: str = "full"):
    r"""Perform convolution of a signal with a linear time-variant system (`ltv`).

    Notes:
        The convolution of two sequences is given by
        $$
        \mathbf{y} = \mathbf{t} \star \mathbf{u}
        $$

        This can be written as a matrix-vector multiplication
        $$
        \mathbf{y} = \mathbf{T} \cdot \mathbf{u}
        $$

        where $T$ is a Toeplitz matrix in which each column represents an impulse response.
        In the case of a linear time-invariant (LTI) system, each column represents a time-shifted copy of the first column.
        In the time-variant case (LTV), every column can contain a unique impulse response, both in values as in size.

        This function assumes all impulse responses are of the same size.
        The input matrix `ltv` thus represents the non-shifted version of the Toeplitz matrix.

    Args:
        signal: Vector representing the input signal $u$.
        ltv: 2D array where each column represents an impulse response.
        mode: {'full', 'valid', 'same'}. Determines the size of the output.

    Returns:
        The result of the convolution operation.

    Raises:
        AssertionError: If the length of the signal does not match the number of columns in `ltv`.

    See Also:
        For convolution with LTI systems.

        - [`np.convolve`][numpy.convolve]
        - [`scipy.signal.convolve`][scipy.signal.convolve]
        - [`scipy.signal.fftconvolve`][scipy.signal.fftconvolve]
    """
    assert len(signal) == ltv.shape[1]

    n = ltv.shape[0] + len(signal) - 1  # Length of output vector
    un = np.concatenate((signal, np.zeros(ltv.shape[0] - 1)))  # Resize input vector
    offsets = np.arange(0, -ltv.shape[0], -1)  # Offsets for impulse responses
    Cs = spdiags(ltv, offsets, n, n)  # Sparse representation of IR's.
    out = Cs.dot(un)  # Calculate dot product.

    if mode == "full":
        return out
    elif mode == "same":
        start = ltv.shape[0] / 2 - 1 + ltv.shape[0] % 2
        stop = len(signal) + ltv.shape[0] / 2 - 1 + ltv.shape[0] % 2
        return out[start:stop]
    elif mode == "valid":
        # length = len(signal) - ltv.shape[0]
        start = ltv.shape[0] - 1
        stop = len(signal)
        return out[start:stop]


def ir2fr(ir, fs, N: int | None = None):
    """Convert impulse response into frequency response. Returns single-sided RMS spectrum.

    Calculates the positive frequencies using [`np.fft.rfft`][numpy.fft.rfft].
    Corrections are then applied to obtain the single-sided spectrum.

    Note:
        Single-sided spectrum. Therefor the amount of bins returned is either N/2 or N/2+1.

    Args:
        ir: Impulse response.
        fs: Sample frequency.
        N: Blocks

    Returns:
        Frequencies and frequency response.

    See Also:
        [`np.fft.rfft`][numpy.fft.rfft]: Used for calculating the positive frequencies.
    """
    N = N if N else ir.shape[-1]
    fr = rfft(ir, n=N) / N
    f = np.fft.rfftfreq(N, 1.0 / fs)  # / 2.0

    fr *= 2.0
    fr[..., 0] /= 2.0  # DC component should not be doubled.
    if not N % 2:  # if not uneven
        fr[..., -1] /= 2.0  # And neither should fs/2 be.

    return f, fr


def decibel_to_neper(decibel):
    r"""Convert decibel to neper.

    Note:
        The conversion is given by
        $$
        \mathrm{dB} = \frac{\log{10}}{20} \mathrm{Np}
        $$

    Args:
        decibel: Value in decibel (dB).

    Returns:
        Value in neper (Np).

    See Also:
        [`neper_to_decibel`][acoustic_toolbox.signal.neper_to_decibel]: For the reverse conversion.
    """
    return np.log(10.0) / 20.0 * decibel


def neper_to_decibel(neper):
    r"""Convert neper to decibel.

    Note:
        The conversion is given by
        $$
        \mathrm{Np} = \frac{20}{\log{10}} \mathrm{dB}
        $$

    Args:
        neper: Value in neper (Np).

    Returns:
        Value in decibel (dB).

    See Also:
        [`decibel_to_neper`][acoustic_toolbox.signal.decibel_to_neper]: For the reverse conversion.
    """
    return 20.0 / np.log(10.0) * neper


class Frequencies:
    """Object describing frequency bands.

    Attributes:
        center: Center frequencies.
        lower: Lower frequencies.
        upper: Upper frequencies.
        bandwidth: Bandwidth.
    """

    center: NDArray[np.float64]
    lower: NDArray[np.float64]
    upper: NDArray[np.float64]
    bandwidth: NDArray[np.float64]

    def __init__(
        self,
        center: NDArray[np.float64] | list[float],
        lower: NDArray[np.float64] | list[float],
        upper: NDArray[np.float64] | list[float],
        bandwidth: NDArray[np.float64] | list[float] | None = None,
    ):
        self.center = np.asarray(center)
        self.lower = np.asarray(lower)
        self.upper = np.asarray(upper)
        self.bandwidth = (
            np.asarray(bandwidth)
            if bandwidth is not None
            else np.asarray(self.upper) - np.asarray(self.lower)
        )

    def __iter__(self):
        for i in range(len(self.center)):
            yield self[i]

    def __len__(self):
        return len(self.center)

    def __str__(self):
        return str(self.center)

    def __repr__(self):
        return "Frequencies({})".format(str(self.center))

    def _get_scalar(self, arr: NDArray[np.float64]) -> float | NDArray[np.float64]:
        """Safely extract a scalar value from a single-element array.

        Args:
            arr: Array to extract scalar from.

        Returns:
            Scalar value if array has one element, otherwise the original array.
        """
        try:
            return arr.item() if arr.size == 1 else arr
        except ValueError:
            return arr  # Return original array if it has multiple elements

    def angular(self):
        """Angular center frequency in radians per second.

        Returns:
            Angular frequencies.
        """
        return 2.0 * np.pi * self.center


class EqualBand(Frequencies):
    """Equal bandwidth spectrum. Generally used for narrowband data.

    Attributes:
        center: Center frequencies.
        fstart: First center frequency.
        fstop: Last center frequency.
        nbands: Amount of frequency bands.
        bandwidth: Bandwidth of bands.
    """

    def __init__(
        self,
        center=None,
        fstart=None,
        fstop=None,
        nbands: int | None = None,
        bandwidth=None,
    ):
        """Equal bandwidth spectrum.

        Raises:
            ValueError: If the center frequencies are not equally spaced.
            ValueError: If insufficient parameters are provided.
        """
        if center is not None:
            try:
                nbands = len(center)
            except TypeError:
                center = [center]
                nbands = 1

            u = np.unique(np.diff(center).round(decimals=3))
            n = len(u)
            if n == 1:
                bandwidth = u
            elif n > 1:
                raise ValueError("Given center frequencies are not equally spaced.")
            else:
                pass
            fstart = center[0]  # - bandwidth/2.0
            fstop = center[-1]  # + bandwidth/2.0
        elif fstart is not None and fstop is not None and nbands:
            bandwidth = (fstop - fstart) / (nbands - 1)
        elif fstart is not None and fstop is not None and bandwidth:
            nbands = round((fstop - fstart) / bandwidth) + 1
        elif fstart is not None and bandwidth and nbands:
            fstop = fstart + nbands * bandwidth
        elif fstop is not None and bandwidth and nbands:
            fstart = fstop - (nbands - 1) * bandwidth
        else:
            raise ValueError(
                "Insufficient parameters. Cannot determine fstart, fstop, bandwidth."
            )

        center = fstart + np.arange(0, nbands) * bandwidth  # + bandwidth/2.0
        upper = fstart + np.arange(0, nbands) * bandwidth + bandwidth / 2.0
        lower = fstart + np.arange(0, nbands) * bandwidth - bandwidth / 2.0

        super(EqualBand, self).__init__(center, lower, upper, bandwidth)

    def __getitem__(self, key):
        return type(self)(center=self.center[key], bandwidth=self.bandwidth)

    def __repr__(self):
        return "EqualBand({})".format(str(self.center))


class OctaveBand(Frequencies):
    """Fractional-octave band spectrum.

    Attributes:
        center: Center frequencies.
        fstart: First center frequency.
        fstop: Last center frequency.
        nbands: Amount of frequency bands.
        bandwidth: Bandwidth.
        fraction: Fraction of fractional-octave filter.
        reference: Reference center frequency.
        nominal: Nominal center frequencies.
    """

    def __init__(
        self,
        center=None,
        fstart=None,
        fstop=None,
        nbands: int | None = None,
        fraction=1,
        reference=acoustic_toolbox.octave.REFERENCE,
    ):
        """Fractional-octave band spectrum.

        Raises:
            ValueError: If insufficient parameters are provided.
        """
        if center is not None:
            try:
                nbands = len(center)
            except TypeError:
                center = [center]
            center = np.asarray(center)
            indices = acoustic_toolbox.octave.index_of_frequency(
                center, fraction=fraction, ref=reference
            )
        elif fstart is not None and fstop is not None:
            nstart = acoustic_toolbox.octave.index_of_frequency(
                fstart, fraction=fraction, ref=reference
            )
            nstop = acoustic_toolbox.octave.index_of_frequency(
                fstop, fraction=fraction, ref=reference
            )
            indices = np.arange(nstart, nstop + 1)
        elif fstart is not None and nbands is not None:
            nstart = acoustic_toolbox.octave.index_of_frequency(
                fstart, fraction=fraction, ref=reference
            )
            indices = np.arange(nstart, nstart + nbands)
        elif fstop is not None and nbands is not None:
            nstop = acoustic_toolbox.octave.index_of_frequency(
                fstop, fraction=fraction, ref=reference
            )
            indices = np.arange(nstop - nbands, nstop)
        else:
            raise ValueError(
                "Insufficient parameters. Cannot determine fstart and/or fstop."
            )

        center = acoustic_toolbox.octave.exact_center_frequency(
            None, fraction=fraction, n=indices, ref=reference
        )
        lower = acoustic_toolbox.octave.lower_frequency(center, fraction=fraction)
        upper = acoustic_toolbox.octave.upper_frequency(center, fraction=fraction)
        bandwidth = upper - lower
        nominal = acoustic_toolbox.octave.nominal_center_frequency(
            None, fraction, indices
        )

        super(OctaveBand, self).__init__(center, lower, upper, bandwidth)

        self.fraction = fraction
        self.reference = reference
        self.nominal = nominal

    def __getitem__(self, key):
        return type(self)(
            center=self.center[key], fraction=self.fraction, reference=self.reference
        )

    def __repr__(self):
        return "OctaveBand({})".format(str(self.center))


def ms(x):
    """Mean value of signal `x` squared.

    Args:
        x: Dynamic quantity.

    Returns:
        Mean squared of `x`.
    """
    return (np.abs(x) ** 2.0).mean()


def rms(x):
    r"""Root mean squared of signal `x`.

    Args:
        x: Dynamic quantity.

    Returns:
        Root mean squared value of `x`.
        $$
        x_{rms} = \lim_{T \to \infty} \sqrt{\frac{1}{T} \int_0^T |f(x)|^2 \mathrm{d} t }
        $$

    See Also:
        [`ms`][acoustic_toolbox.signal.ms]
    """
    return np.sqrt(ms(x))


def normalize(y, x=None):
    r"""Normalize power in `y` to a (standard normal) white noise signal.

    Optionally normalize to power in signal `x`.

    Note:
        The mean power of a Gaussian with $\mu=0$ and $\sigma=1$ is 1.

    Args:
        y: Signal to be normalized.
        x: Reference signal. Defaults to None.

    Returns:
        Normalized signal.
    """
    if x is not None:
        x = ms(x)
    else:
        x = 1.0
    return y * np.sqrt(x / ms(y))
    # return y * np.sqrt( 1.0 / (np.abs(y)**2.0).mean() )

    ## Broken? Caused correlation in auralizations....weird!


def window_scaling_factor(window, axis: int = -1):
    r"""Calculate window scaling factor.

    When analysing broadband (filtered noise) signals, it is common to normalize
    the windowed signal so that it has the same power as the un-windowed signal.
    $$
    S = \sqrt{\frac{\sum_{i=0}^N w_i^2}{N}}
    $$

    Args:
        window: Window.
        axis: Axis along which to calculate.

    Returns:
        Window scaling factor.
    """
    return np.sqrt((window * window).mean(axis=axis))


def apply_window(x, window):
    """Apply window to signal.

    $$
    x_s(t) = x(t) / S
    $$

    where $S$ is the window scaling factor.

    Args:
        x: Instantaneous signal $x(t)$.
        window: Vector representing window.

    Returns:
        Signal with window applied.

    See Also:
        [`window_scaling_factor`][acoustic_toolbox.signal.window_scaling_factor]: For calculating the scaling factor.
    """
    s = window_scaling_factor(window)  # Determine window scaling factor.
    n = len(window)
    windows = x // n  # Amount of windows.
    x = x[0 : windows * n]  # Truncate final part of signal that does not fit.
    # x = x.reshape(-1, len(window)) # Reshape so we can apply window.
    y = np.tile(window, windows)

    return x * y / s


def amplitude_spectrum(x, fs, N: int | None = None):
    r"""Amplitude spectrum of instantaneous signal $x(t)$.

    The amplitude spectrum  gives the amplitudes of the sinusoidal the signal is built
    up from, and the RMS (root-mean-square) amplitudes can easily be found by dividing
    these amplitudes with $\sqrt{2}$

    The amplitude spectrum is double-sided.

    Args:
        x: Instantaneous signal $x(t)$.
        fs: Sample frequency $f_s$.
        N: Number of FFT bins.

    Returns:
        tuple: Frequencies and amplitude spectrum.
    """
    N = N if N else x.shape[-1]
    fr = np.fft.fft(x, n=N) / N
    f = np.fft.fftfreq(N, 1.0 / fs)
    return np.fft.fftshift(f), np.fft.fftshift(fr, axes=[-1])


def auto_spectrum(x, fs, N: int | None = None):
    r"""Auto-spectrum of instantaneous signal $x(t)$.

    The auto-spectrum contains the squared amplitudes of the signal. Squared amplitudes
    are used when presenting data as it is a measure of the power/energy in the signal.

    $$
    S_{xx} (f_n) = \overline{X (f_n)} \cdot X (f_n)
    $$

    The auto-spectrum is double-sided.

    Args:
        x: Instantaneous signal $x(t)$.
        fs: Sample frequency $f_s$.
        N: Number of FFT bins.

    Returns:
        f: Frequencies
        a: Auto-spectrum
    """
    f, a = amplitude_spectrum(x, fs, N=N)
    return f, (a * a.conj()).real


def power_spectrum(x, fs, N: int | None = None):
    r"""Power spectrum of instantaneous signal $x(t)$.


    The power spectrum, or single-sided autospectrum, contains the squared RMS amplitudes of the signal.

    A power spectrum is a spectrum with squared RMS values. The power spectrum is
    calculated from the autospectrum of the signal.

    Warning:
        Does not include scaling to reference value!

    Args:
        x: Instantaneous signal $x(t)$.
        fs: Sample frequency $f_s$.
        N: Number of FFT bins.

    Returns:
        f: Frequencies
        a: Power spectrum

    See Also:
        [`auto_spectrum`][acoustic_toolbox.signal.auto_spectrum]
    """
    N = N if N else x.shape[-1]
    f, a = auto_spectrum(x, fs, N=N)
    a = a[..., N // 2 :]
    f = f[..., N // 2 :]
    a *= 2.0
    a[..., 0] /= 2.0  # DC component should not be doubled.
    if not N % 2:  # if not uneven
        a[..., -1] /= 2.0  # And neither should fs/2 be.
    return f, a


def angle_spectrum(x, fs, N: int | None = None):
    r"""Phase angle spectrum of instantaneous signal $x(t)$.

    This function returns a single-sided phase angle spectrum.

    Args:
        x: Instantaneous signal $x(t)$.
        fs: Sample frequency $f_s$.
        N: Number of FFT bins.

    Returns:
        f: Frequencies
        a: Phase angle spectrum

    See Also:
        [`phase_spectrum`][acoustic_toolbox.signal.phase_spectrum]: For unwrapped phase spectrum.
    """
    N = N if N else x.shape[-1]
    f, a = amplitude_spectrum(x, fs, N)
    a = np.angle(a)
    a = a[..., N // 2 :]
    f = f[..., N // 2 :]
    return f, a


def phase_spectrum(x, fs, N: int | None = None):
    r"""Phase spectrum of instantaneous signal $x(t)$.

    This function returns single-sided unwrapped phase spectrum.

    Args:
        x: Instantaneous signal $x(t)$.
        fs: Sample frequency $f_s$.
        N: Number of FFT bins.

    Returns:
        f: Frequencies
        a: Unwrapped phase spectrum

    See Also:
        [`angle_spectrum`][acoustic_toolbox.signal.angle_spectrum]: For wrapped phase angle.
    """
    f, a = angle_spectrum(x, fs, N=None)
    return f, np.unwrap(a)


def density_spectrum(x, fs, N: int | None = None):
    """Density spectrum of instantaneous signal $x(t)$.

    A density spectrum considers the amplitudes per unit frequency.
    Density spectra are used to compare spectra with different frequency resolution as the
    magnitudes are not influenced by the resolution because it is per Hertz. The amplitude
    spectra on the other hand depend on the chosen frequency resolution.

    Args:
        x: Instantaneous signal $x(t)$.
        fs: Sample frequency $f_s$.
        N: Number of FFT bins.

    Returns:
        f: Frequencies
        a: Density spectrum
    """
    N = N if N else x.shape[-1]
    fr = np.fft.fft(x, n=N) / fs
    f = np.fft.fftfreq(N, 1.0 / fs)
    return np.fft.fftshift(f), np.fft.fftshift(fr)


def integrate_bands(data, a, b):
    """Reduce frequency resolution of power spectrum. Merges frequency bands by integration.


    Args:
        data: Vector with narrowband powers.
        a: Instance of `Frequencies`.
        b: Instance of `Frequencies`.

    Returns:
        Integrated bands.

    Raises:
        NotImplementedError: If the ratio of fractional-octaves is not an integer.

    Todo:
        Needs rewriting so that the summation goes over axis=1.
    """
    try:
        if b.fraction % a.fraction:
            raise NotImplementedError(
                "Non-integer ratio of fractional-octaves are not supported."
            )
    except AttributeError:
        pass

    lower, _ = np.meshgrid(b.lower, a.center)
    upper, _ = np.meshgrid(b.upper, a.center)
    _, center = np.meshgrid(b.center, a.center)

    return ((lower < center) * (center <= upper) * data[..., None]).sum(axis=-2)


def bandpass_frequencies(
    x,
    fs,
    frequencies: Frequencies,
    order: int = 8,
    purge: bool = False,
    zero_phase: bool = False,
) -> tuple[Frequencies, np.ndarray]:
    """Apply bandpass filters for frequencies.

    Args:
        x: Instantaneous signal $x(t)$.
        fs: Sample frequency.
        frequencies: Instance of `Frequencies`.
        order: Filter order.
        purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        zero_phase: Prevent phase error by filtering in both directions (filtfilt)

    Returns:
        OctaveBand: Instance of `OctaveBand`
        np.ndarray: Filtered array.
    """
    if purge:
        frequencies = frequencies[frequencies.upper < fs / 2.0]
    return frequencies, np.array(
        [
            bandpass(
                x,
                band._get_scalar(band.lower),
                band._get_scalar(band.upper),
                fs,
                order,
                zero_phase=zero_phase,
            )
            for band in frequencies
        ]
    )


def bandpass_octaves(
    x,
    fs,
    frequencies=NOMINAL_OCTAVE_CENTER_FREQUENCIES,
    order=8,
    purge=False,
    zero_phase: bool = False,
) -> tuple[OctaveBand, np.ndarray]:
    """Apply 1/1-octave bandpass filters.

    Args:
        x: Instantaneous signal $x(t)$.
        fs: Sample frequency.
        frequencies: Center frequencies.
        order: Filter order.
        purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        zero_phase: Prevent phase error by filtering in both directions (filtfilt)

    Returns:
        OctaveBand: Instance of `OctaveBand`
        np.ndarray: Filtered array.

    See Also:
        [`octavepass`][acoustic_toolbox.signal.octavepass]
    """
    return bandpass_fractional_octaves(
        x, fs, frequencies, fraction=1, order=order, purge=purge, zero_phase=zero_phase
    )


def bandpass_third_octaves(
    x,
    fs,
    frequencies=NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES,
    order: int = 8,
    purge: bool = False,
    zero_phase: bool = False,
) -> tuple[OctaveBand, np.ndarray]:
    """Apply 1/3-octave bandpass filters.

    Args:
        x: Instantaneous signal $x(t)$.
        fs: Sample frequency.
        frequencies: Center frequencies.
        order: Filter order.
        purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        zero_phase: Prevent phase error by filtering in both directions (filtfilt)

    Returns:
        OctaveBand: Instance of `OctaveBand`
        np.ndarray: Filtered array.

    See Also:
        [`octavepass`][acoustic_toolbox.signal.octavepass]
    """
    return bandpass_fractional_octaves(
        x, fs, frequencies, fraction=3, order=order, purge=purge, zero_phase=zero_phase
    )


def bandpass_fractional_octaves(
    x,
    fs,
    frequencies,
    fraction=None,
    order: int = 8,
    purge: bool = False,
    zero_phase: bool = False,
) -> tuple[OctaveBand, np.ndarray]:
    """Apply 1/N-octave bandpass filters.

    Args:
        x: Instantaneous signal $x(t)$.
        fs: Sample frequency.
        frequencies: Center frequencies or instance of `OctaveBand`.
        fraction: Fraction of fractional-octave band.
        order: Filter order.
        purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        zero_phase: Prevent phase error by filtering in both directions (filtfilt)

    Returns:
        OctaveBand: Instance of `OctaveBand`
        np.ndarray: Filtered array.

    See Also:
        [`octavepass`][acoustic_toolbox.signal.octavepass]
    """
    if not isinstance(frequencies, Frequencies):
        frequencies = OctaveBand(center=frequencies, fraction=fraction)
    return bandpass_frequencies(
        x, fs, frequencies, order=order, purge=purge, zero_phase=zero_phase
    )


def third_octaves(
    p,
    fs,
    density: bool = False,
    frequencies=NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES,
    ref: float = REFERENCE_PRESSURE,
) -> tuple[OctaveBand, np.ndarray]:
    """Calculate level per 1/3-octave in frequency domain using the FFT.

    Note:
        Exact center frequencies are always calculated.

    Args:
        p: Instantaneous signal $x(t)$.
        fs: Sample frequency.
        density: Calculate power density instead of power.
        frequencies: Center frequencies.
        ref: Reference pressure.

    Returns:
        OctaveBand: Instance of `OctaveBand`
        np.ndarray: Level array.

    See Also:
        [`NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES`][acoustic_toolbox.bands.NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES]
    """
    fob = OctaveBand(center=frequencies, fraction=3)
    f, p = power_spectrum(p, fs)
    fnb = EqualBand(f)
    power = integrate_bands(p, fnb, fob)
    if density:
        power /= fob.bandwidth / fnb.bandwidth
    level = 10.0 * np.log10(power / ref**2.0)
    return fob, level


def octaves(
    p,
    fs,
    density=False,
    frequencies=NOMINAL_OCTAVE_CENTER_FREQUENCIES,
    ref=REFERENCE_PRESSURE,
):
    """Calculate level per 1/1-octave in frequency domain using the FFT.

    Notes:
        - Based on power spectrum (FFT)
        - Exact center frequencies are always calculated.

    Args:
        p: Instantaneous signal $x(t)$.
        fs: Sample frequency.
        density: Calculate power density instead of power.
        frequencies: Center frequencies.
        ref: Reference value.

    Returns:
        OctaveBand: Instance of `OctaveBand`
        np.ndarray: Level array.

    See Also:
        [`NOMINAL_OCTAVE_CENTER_FREQUENCIES`][acoustic_toolbox.bands.NOMINAL_OCTAVE_CENTER_FREQUENCIES]
    """
    fob = OctaveBand(center=frequencies, fraction=1)
    f, p = power_spectrum(p, fs)
    fnb = EqualBand(f)
    power = integrate_bands(p, fnb, fob)
    if density:
        power /= fob.bandwidth / fnb.bandwidth
    level = 10.0 * np.log10(power / ref**2.0)
    return fob, level


def fractional_octaves(
    p,
    fs,
    start: float = 5.0,
    stop: float = 16000.0,
    fraction: int = 3,
    density: bool = False,
    ref: float = REFERENCE_PRESSURE,
):
    """Calculate level per 1/N-octave in frequency domain using the FFT. N is `fraction`.

    Notes:
        - Based on power spectrum (FFT)
        - This function does *not* use nominal center frequencies.
        - Exact center frequencies are always calculated.

    Args:
        p: Instantaneous signal $x(t)$.
        fs: Sample frequency.
        start: Start frequency.
        stop: Stop frequency.
        fraction: Fraction of fractional-octave band.
        density: Calculate power density instead of power.
        ref: Reference value.

    Returns:
        OctaveBand: Instance of `OctaveBand`
        np.ndarray: Level array.
    """
    fob = OctaveBand(fstart=start, fstop=stop, fraction=fraction)
    f, p = power_spectrum(p, fs)
    fnb = EqualBand(f)
    power = integrate_bands(p, fnb, fob)
    if density:
        power /= fob.bandwidth / fnb.bandwidth
    level = 10.0 * np.log10(power / ref**2.0)
    return fob, level


class Filterbank:
    """Fractional-Octave filter bank.

    Warning:
        For high frequencies the filter coefficients are wrong for low frequencies.
        Therefore, to improve the response for lower frequencies the signal should be downsampled.
        Currently, there is no easy way to do so within the Filterbank.

    Attributes:
        frequencies: Frequencies object.
            See also [`Frequencies`][acoustic_toolbox.signal.Frequencies] and subclasses.

            **Note:** A frequencies attribute should have the attributes center, lower, and upper.

        order: Filter order of Butterworth filter.
        sample_frequency: Sample frequency.
    """

    def __init__(
        self,
        frequencies: Frequencies,
        sample_frequency: float = 44100,
        order: int = 8,
    ):
        self.frequencies = frequencies
        self.order = order
        self.sample_frequency = sample_frequency

    @property
    def sample_frequency(self):
        """Sample frequency.

        Returns:
            float: Sample frequency.
        """
        return self._sample_frequency

    @sample_frequency.setter
    def sample_frequency(self, x):
        self._sample_frequency = x

    @property
    def filters(self):
        """Filters this filterbank consists of.

        Returns:
            generator: Filter coefficients for each band.
        """
        fs = self.sample_frequency
        return (
            bandpass_filter(
                self.frequencies._get_scalar(lower),
                self.frequencies._get_scalar(upper),
                fs,
                order=self.order,
                output="sos",
            )
            for lower, upper in zip(self.frequencies.lower, self.frequencies.upper)
        )

        # order = self.order
        # filters = list()
        # nyq = self.sample_frequency / 2.0
        # return ( butter(order, [lower/nyq, upper/nyq], btype='band', analog=False) for lower, upper in zip(self.frequencies.lower, self.frequencies.upper) )

    def lfilter(
        self, signal: NDArray[np.float64]
    ) -> Generator[NDArray[np.float64], None, None]:
        """Filter signal with filterbank.

        Note:
            This function uses [`scipy.signal.lfilter`][scipy.signal.lfilter].

        Args:
            signal: Signal to be filtered.

        Returns:
            generator: Filtered signal for each band.
        """
        return (sosfilt(sos, signal) for sos in self.filters)

    def filtfilt(
        self, signal: NDArray[np.float64]
    ) -> Generator[NDArray[np.float64], None, None]:
        """Filter signal with filterbank.

        Note:
            This function uses [`scipy.signal.filtfilt`][scipy.signal.filtfilt] and therefore has a zero-phase response.

        Args:
            signal: Signal to be filtered.

        Returns:
            List consisting of a filtered signal per filter.
        """
        return (_sosfiltfilt(sos, signal) for sos in self.filters)

    def power(self, signal: NDArray[np.float64]) -> NDArray[np.float64]:
        """Power per band in signal.

        Args:
            signal: Signal to be analyzed.

        Returns:
            np.ndarray: Power per band.
        """
        filtered = self.filtfilt(signal)
        return np.array(
            [
                (x**2.0).sum() / len(x) / bw
                for x, bw in zip(filtered, self.frequencies.bandwidth)
            ]
        )

    def plot_response(self):
        """Plot frequency response.

        Note:
            The following phase response is obtained in case [`lfilter`][acoustic_toolbox.signal.Filterbank.lfilter] is used.
            The method [`filtfilt`][acoustic_toolbox.signal.Filterbank.filtfilt] has a zero-phase response.

        Returns:
            matplotlib.figure.Figure: Figure with frequency response plot.
        """
        fs = self.sample_frequency
        fig = plt.figure()
        ax1 = fig.add_subplot(211)
        ax2 = fig.add_subplot(212)
        for f, fc in zip(self.filters, self.frequencies.center):
            w, h = freqz(f[0], f[1], int(fs / 2))  # np.arange(fs/2.0))
            ax1.semilogx(
                w / (2.0 * np.pi) * fs, 20.0 * np.log10(np.abs(h)), label=str(int(fc))
            )
            ax2.semilogx(w / (2.0 * np.pi) * fs, np.angle(h), label=str(int(fc)))
        ax1.set_xlabel(r"$f$ in Hz")
        ax1.set_ylabel(r"$|H|$ in dB re. 1")
        ax2.set_xlabel(r"$f$ in Hz")
        ax2.set_ylabel(r"$\angle H$ in rad")
        ax1.legend(loc=5)
        ax2.legend(loc=5)
        ax1.set_ylim(-60.0, +10.0)

        return fig

    def plot_power(self, signal):
        """Plot power in signal.

        Args:
            signal: Signal to be analyzed.

        Returns:
            matplotlib.figure.Figure: Figure with power plot.
        """
        f = self.frequencies.center
        p = self.power(signal)

        fig = plt.figure()
        ax = fig.add_subplot(111)
        p = ax.bar(f, 20.0 * np.log10(p))
        ax.set_xlabel("$f$ in Hz")
        ax.set_ylabel("$L$ in dB re. 1")
        ax.set_xscale("log")

        return fig


def isolate(signals):
    """Isolate signals using Singular Value Decomposition.

    Args:
        signals: Array of shape N x M where N is the amount of samples and M the amount of signals. Thus, each column is a signal

    Returns:
        Array of isolated signals.
    """
    x = np.asarray(signals)

    W, s, v = np.linalg.svd((np.tile((x * x).sum(axis=0), (len(x), 1)) * x).dot(x.T))
    return v.T


def zero_crossings(data):
    """Determine the positions of zero crossings in `data`.

    Args:
        data: Vector.

    Returns:
        Vector with indices of samples *before* the zero crossing.
    """
    pos = data > 0
    npos = ~pos
    return ((pos[:-1] & npos[1:]) | (npos[:-1] & pos[1:])).nonzero()[0]


def amplitude_envelope(signal: np.ndarray, fs, axis=-1):
    """Instantaneous amplitude of tone.

    The instantaneous amplitude is the magnitude of the analytic signal.

    Args:
        signal: Signal.
        fs: Sample frequency.
        axis: Axis. Defaults to -1.

    Returns:
        Amplitude envelope of `signal`.

    See Also:
        [`hilbert`][scipy.signal.hilbert]
    """
    return np.abs(hilbert(signal, axis=axis))


def instantaneous_phase(signal: np.ndarray, fs, axis=-1):
    """Instantaneous phase of tone.

    The instantaneous phase is the angle of the analytic signal.
    This function returns a wrapped angle.

    Args:
        signal: Signal.
        fs: Sample frequency.
        axis: Axis.

    Returns:
        Instantaneous phase of `signal`.

    See Also:
        [`hilbert`][scipy.signal.hilbert]
    """
    return np.angle(hilbert(signal, axis=axis))


def instantaneous_frequency(signal: np.ndarray, fs, axis=-1):
    """Determine instantaneous frequency of tone.

    The instantaneous frequency can be obtained by differentiating the unwrapped instantaneous phase.

    Args:
        signal: Signal.
        fs: Sample frequency.
        axis: Axis.

    Returns:
        Instantaneous frequency of `signal`.

    See Also:
        [`instantaneous_phase`][acoustic_toolbox.signal.instantaneous_phase]
    """
    return (
        np.diff(
            np.unwrap(instantaneous_phase(signal, fs, axis=axis), axis=axis), axis=axis
        )
        / (2.0 * np.pi)
        * fs
    )


def wvd(signal: np.ndarray, fs, analytic=True):
    r"""Wigner-Ville Distribution.

    $$
    W_z(n, \omega) = 2 \sum_k z^*[n-k]z[n+k] e^{-j\omega 2kT}
    $$

    Includes positive and negative frequencies.

    Args:
        signal: Signal.
        fs: Sample frequency.
        analytic: If True, uses the analytic signal. Defaults to True.

    Returns:
        Frequencies: Instance of `Frequencies`.
        W.T: Wigner-Ville distribution
    """
    signal = np.asarray(signal)

    N = int(len(signal) + len(signal) % 2)
    length_FFT = N  # Take an even value of N

    length_time = len(signal)

    if analytic:
        signal = hilbert(signal)
    s = np.concatenate((np.zeros(length_time), signal, np.zeros(length_time)))
    W = np.zeros((length_FFT, length_time))
    tau = np.arange(0, N // 2)

    R = np.zeros((N, length_time), dtype="float64")

    i = length_time
    for t in range(length_time):
        R[t, tau] = s[i + tau] * s[i - tau].conj()  # In one direction
        R[t, N - (tau + 1)] = R[t, tau + 1].conj()  # And the other direction
        i += 1
    W = np.fft.fft(R, length_FFT) / (2 * length_FFT)

    f = np.fft.fftfreq(N, 1.0 / fs)
    return f, W.T


def _sosfiltfilt(sos, x, axis=-1, padtype="odd", padlen=None, method="pad", irlen=None):
    """Filtfilt version using Second Order sections.

    Code is taken from [scipy.signal.sosfiltfilt](https://github.com/scipy/scipy/blob/main/scipy/signal/_filter_design.py#L1100)
    and adapted to make it work with SOS.

    Note that boradcasting does not work.
    """
    from scipy.signal import sosfilt_zi
    from scipy.signal._arraytools import axis_slice, axis_reverse

    x = np.asarray(x)

    if padlen is None:
        edge = 0
    else:
        edge = padlen

    # x's 'axis' dimension must be bigger than edge.
    if x.shape[axis] <= edge:
        raise ValueError(
            "The length of the input vector x must be at least "
            "padlen, which is %d." % edge
        )

    if padtype is not None and edge > 0:
        # Make an extension of length `edge` at each
        # end of the input array.
        if padtype == "even":
            ext = even_ext(x, edge, axis=axis)
        elif padtype == "odd":
            ext = odd_ext(x, edge, axis=axis)
        else:
            ext = const_ext(x, edge, axis=axis)
    else:
        ext = x

    # Get the steady state of the filter's step response.
    zi = sosfilt_zi(sos)

    # Reshape zi and create x0 so that zi*x0 broadcasts
    # to the correct value for the 'zi' keyword argument
    # to lfilter.
    # zi_shape = [1] * x.ndim
    # zi_shape[axis] = zi.size
    # zi = np.reshape(zi, zi_shape)
    x0 = axis_slice(ext, stop=1, axis=axis)
    # Forward filter.
    (y, zf) = sosfilt(sos, ext, axis=axis, zi=zi * x0)

    # Backward filter.
    # Create y0 so zi*y0 broadcasts appropriately.
    y0 = axis_slice(y, start=-1, axis=axis)
    (y, zf) = sosfilt(sos, axis_reverse(y, axis=axis), axis=axis, zi=zi * y0)

    # Reverse y.
    y = axis_reverse(y, axis=axis)

    if edge > 0:
        # Slice the actual signal from the extended signal.
        y = axis_slice(y, start=edge, stop=-edge, axis=axis)

    return y


def decimate(
    x: np.ndarray,
    q: int,
    n: int | None = None,
    ftype: str = "iir",
    axis: int = -1,
    zero_phase: bool = False,
) -> np.ndarray:
    """Downsample the signal by using a filter.

    By default, an order 8 Chebyshev type I filter is used. A 30 point FIR
    filter with hamming window is used if `ftype` is 'fir'.

    Notes:
        - The `zero_phase` keyword was added in v0.17.0.
        - The possibility to use instances of `lti` as `ftype` was added in v0.17.0

    Args:
        x: The signal to be downsampled.
        q: The downsampling factor.
        n: The order of the filter.
        ftype: The type of the lowpass filter.
        axis: The axis along which to decimate.
        zero_phase: Prevent phase shift by filtering with `filtfilt` instead of `lfilter`.

    Returns:
        The down-sampled signal.

    See Also:
        `resample`: For resampling the signal.
    """
    if not isinstance(q, int):
        raise TypeError("q must be an integer")

    if ftype == "fir":
        if n is None:
            n = 30
        system = lti(firwin(n + 1, 1.0 / q, window="hamming"), 1.0)

    elif ftype == "iir":
        if n is None:
            n = 8
        system = lti(*cheby1(n, 0.05, 0.8 / q))
    else:
        system = ftype

    if zero_phase:
        y = filtfilt(system.num, system.den, x, axis=axis)
    else:
        y = lfilter(system.num, system.den, x, axis=axis)

    sl = [slice(None)] * y.ndim
    sl[axis] = slice(None, None, q)
    return y[tuple(sl)]


def impulse_response_real_even(tf, ntaps):
    """The impulse response of a real and even frequency response is also real and even.

    A symmetric impulse response is needed. The center of symmetry determines the delay
    of the filter and thereby whether the filter is causal (delay>0, linear-phase) or
    non-causal (delay=0, linear-phase, zero-phase).

    Creating linear phase can be done by multiplying the magnitude with a complex
    exponential corresponding to the desired shift. Another method is to rotate the
    impulse response.

    [https://ccrma.stanford.edu/~jos/filters/Zero_Phase_Filters_Even_Impulse.html](https://ccrma.stanford.edu/~jos/filters/Zero_Phase_Filters_Even_Impulse.html)

    Args:
        tf: Real and even frequency response. Only positive frequencies.
        ntaps: Amount of taps.

    Returns:
        A real and even impulse response with length `ntaps`.
    """
    ir = np.fft.ifftshift(np.fft.irfft(tf, n=ntaps)).real
    return ir


def linear_phase(ntaps, steepness=1):
    """Compute linear phase delay for a single-sided spectrum.

    A linear phase delay can be added to an impulse response using the function [`np.fft.ifftshift`][numpy.fft.ifftshift].
    Sometimes, however, you would like to add the linear phase delay to the frequency response instead.
    This function computes the linear phase delay which can be multiplied with a single-sided frequency response.

    Args:
        ntaps: Amount of filter taps.
        steepness: Steepness of phase delay. Default value is 1, corresponding to delay in samples of `ntaps // 2`.

    Returns:
        np.ndarray: Linear phase delay.
    """
    f = np.fft.rfftfreq(ntaps, 1.0)  # Frequencies normalized to Nyquist.
    alpha = ntaps // 2 * steepness
    return np.exp(-1j * 2.0 * np.pi * f * alpha)


__all__ = [
    "bandpass",
    "bandpass_frequencies",
    "bandpass_fractional_octaves",
    "bandpass_octaves",
    "bandpass_third_octaves",
    "lowpass",
    "highpass",
    "octavepass",
    "octave_filter",
    "bandpass_filter",
    "convolve",
    "ir2fr",
    "decibel_to_neper",
    "neper_to_decibel",
    "EqualBand",
    "OctaveBand",
    "ms",
    "rms",
    "normalize",
    "window_scaling_factor",
    "apply_window",
    "amplitude_spectrum",
    "auto_spectrum",
    "power_spectrum",
    "angle_spectrum",
    "phase_spectrum",
    "density_spectrum",
    "integrate_bands",
    "octaves",
    "third_octaves",
    "fractional_octaves",
    "Filterbank",
    "isolate",
    "zero_crossings",
    "amplitude_envelope",
    "instantaneous_phase",
    "instantaneous_frequency",
    "wvd",
    "decimate",
]



from acoustic_toolbox.standards.iso_tr_25417_2007 import *


def test_sound_exposure_level():
    duration = 20.0
    fs = 44100.0
    samples = int(fs * duration)

    signal = np.random.randn(samples)

    leq = equivalent_sound_pressure_level(signal)
    sel = sound_exposure_level(signal, fs)

    assert ((sel - leq) - 10.0 * np.log10(duration)) < 0.1



"""The Quantity module provides two classes to work with quantities and units."""

from acoustic_toolbox.standards.iso_tr_25417_2007 import REFERENCE_PRESSURE

quantities = {
    "pressure": ("Pressure", "pascal", True, "p", "$p$", REFERENCE_PRESSURE),
}
"""Dictionary with quantities. Each quantity is stored as a tuple."""

units = {
    "meter": ("meter", "m", "$m$"),
    "pascal": ("pascal", "Pa", "$Pa$"),
}
"""Dictionary with units. Each unit is stored as a tuple."""


class Unit:
    """Unit of quantity.

    Note:
      Perhaps inherit from tuple or [`collections.namedtuple`][collections.namedtuple]?
    """

    def __init__(self, name: str, symbol: str, symbol_latex: str):
        """Initialize the unit.

        Args:
          name: Name of the unit.
          symbol: Symbol of the unit.
          symbol_latex: Symbol of the unit in LaTeX.
        """
        self.name = name
        """Name of the unit."""

        self.symbol = symbol
        """Symbol of the unit."""

        self.symbol_latex = symbol_latex
        """Symbol of the unit in LaTeX."""

    def __repr__(self):
        return "Unit({})".format(self.name)

    def __str__(self):
        return self.name


class Quantity:
    """Quantity."""

    def __init__(
        self,
        name: str,
        unit: Unit,
        dynamic: bool,
        symbol: str | None = None,
        symbol_latex: str | None = None,
        reference: float = 1.0,
    ):
        """Initialize the quantity.

        Args:
          name: Name of the quantity.
          unit: Unit of the quantity.
          dynamic: Dynamic quantity (`True`) or energetic (`False`).
          symbol: Symbol of the quantity.
          symbol_latex: Symbol of the quantity in LaTeX.
          reference: Reference value of the quantity.
        """
        self.name = name
        """Name of the quantity."""

        self.symbol = symbol
        """Symbol of the quantity."""

        self.symbol_latex = symbol_latex
        """Symbol of the unit in LaTeX."""

        self.unit = unit
        """Unit. See [`Unit`][acoustic_toolbox.quantity.Unit]."""

        self.dynamic = dynamic
        """Dynamic quantity (`True`) or energetic (`False`)."""

        self.reference = reference
        """Reference value of the quantity."""

    def __repr__(self):
        return "Quantity({})".format(self.name)

    def __str__(self):
        return self.name

    @property
    def energetic(self) -> bool:
        """Energetic quantity (`True`) or dynamic (`False`)."""
        return not self.dynamic


def get_quantity(name: str) -> Quantity:
    """Get quantity by name. Returns instance of [`Quantity`][acoustic_toolbox.quantity.Quantity].

    Args:
      name: Name of the quantity.

    Returns:
      Instance of [`Quantity`][acoustic_toolbox.quantity.Quantity].
    """
    try:
        q = list(quantities[name])
    except KeyError:
        raise ValueError("Unknown quantity. Quantity is not yet specified.")
    try:
        q[1] = Unit(*units[name])
    except KeyError:
        raise RuntimeError(
            "Unknown unit. Quantity has been specified but unit has not."
        )

    return Quantity(*q)



import numpy as np
from acoustic_toolbox.bands import octave, third
from acoustic_toolbox.imaging import plot_octave, plot_third, plot_bands


class TestImaging:
    """Test :mod:`acoustic_toolbox.imaging`"""

    octaves = octave(16, 16000)
    thirds = third(63, 8000)
    tl_oct = np.array([3, 4, 5, 12, 15, 24, 28, 23, 35, 45, 55])
    tl_third = np.array(
        [0, 0, 0, 1, 1, 2, 3, 5, 8, 13, 21, 32, 41, 47, 46, 44, 58, 77, 61, 75, 56, 54]
    )
    title = "Title"
    label = "Label"

    def test_plot_octave(self):
        plot_octave(self.tl_oct, self.octaves)

    def test_plot_octave_kHz(self):
        plot_octave(
            self.tl_oct,
            self.octaves,
            kHz=True,
            xlabel=self.label,
            ylabel=self.label,
            title=self.title,
            separator=".",
        )

    def test_plot_third_octave(self):
        plot_third(self.tl_third, self.thirds, marker="s", separator=",")

    def test_plot_third_octave_kHz(self):
        plot_third(
            self.tl_third,
            self.thirds,
            marker="s",
            kHz=True,
            xlabel=self.label,
            ylabel=self.label,
            title=self.title,
        )

    def test_plot_band_oct(self):
        plot_bands(self.tl_oct, self.octaves, axes=None, band_type="octave")



# Ambisonics

import numpy as np
import scipy.special
from typing import Generator, Tuple
from numpy.typing import NDArray


def acn(order: int = 1) -> Generator[Tuple[int, int], None, None]:
    """Spherical harmonic degree `n` and order `m` for ambisonics order `order`.

    Follows ACN.

    | ACN | n | m  | letter |
    |-----|---|----|--------|
    | 0   | 0 | 0  | W      |
    | 1   | 1 | -1 | Y      |
    | 2   | 1 | 0  | Z      |
    | 3   | 1 | +1 | X      |

    Args:
        order: Ambisonics order.

    Yields:
        Tuple[int, int]: Degree `n` and order `m`.
    """
    for n in range(order + 1):
        for m in range(-n, n + 1):
            yield (n, m)


def sn3d(m: int | NDArray[np.int_], n: int | NDArray[np.int_]) -> NDArray[np.float64]:
    """SN3D or Schmidt semi-normalisation.

    See [http://en.wikipedia.org/wiki/Ambisonic_data_exchange_formats#SN3D](http://en.wikipedia.org/wiki/Ambisonic_data_exchange_formats#SN3D)

    Args:
        m: Order `n`
        n: Degree `m`

    Returns:
        NDArray[np.float64]: The SN3D normalization coefficients.
    """
    m = np.atleast_1d(m)
    n = np.atleast_1d(n)

    d = np.logical_not(m.astype(bool))
    out = np.sqrt(
        (2.0 - d)
        / (4.0 * np.pi)
        * scipy.special.factorial(n - np.abs(m))
        / scipy.special.factorial.factorial(n + np.abs(m))
    )
    return out


def n3d(m: int | NDArray[np.int_], n: int | NDArray[np.int_]) -> NDArray[np.float64]:
    """N3D or full three-D normalisation.

    See [http://en.wikipedia.org/wiki/Ambisonic_data_exchange_formats#N3D](http://en.wikipedia.org/wiki/Ambisonic_data_exchange_formats#N3D)

    Args:
        m: Order `n`
        n: Degree `m`

    Returns:
        NDArray[np.float64]: The N3D normalization coefficients.
    """
    n = np.atleast_1d(n)
    return sn3d(m, n) * np.sqrt(2 * n + 1)


__all__ = ["acn", "sn3d", "n3d"]



"""The building module contains functions for calculating acoustic properties of building elements, including
sound transmission class (STC) and weighted sound reduction index ($R_w$) calculations.

Functions:
    rw_curve: Calculate the reference curve for weighted sound reduction index.
    rw: Calculate the weighted sound reduction index ($R_w$).
    rw_c: Calculate the weighted sound reduction index with spectrum adaptation term C ($R_w + C$).
    rw_ctr: Calculate the weighted sound reduction index with spectrum adaptation term $C_{tr}$ ($R_w + C_{tr}$).
    stc_curve: Calculate the Sound Transmission Class (STC) reference curve.
    stc: Calculate the Sound Transmission Class (STC).
    mass_law: Calculate transmission loss according to mass law.

"""

import numpy as np
from numpy.typing import NDArray


def rw_curve(tl: NDArray[np.float64]) -> NDArray[np.float64]:
    """Calculate the curve of $Rw$ from a NumPy array `tl` with third
    octave data between 100 Hz and 3.15 kHz.

    Args:
        tl: Transmission Loss

    Returns:
        NDArray[np.float64]: Reference curve values.
    """
    ref_curve = np.array([0, 3, 6, 9, 12, 15, 18, 19, 20, 21, 22, 23, 23, 23, 23, 23])
    residuals = 0
    while residuals > -32:
        ref_curve += 1
        diff = tl - ref_curve
        residuals = np.sum(np.clip(diff, np.min(diff), 0))
    ref_curve -= 1
    return ref_curve


def rw(tl: NDArray[np.float64]) -> float:
    """Calculate $R_W$ from a NumPy array `tl` with third octave data
    between 100 Hz and 3.15 kHz.

    Args:
        tl: Transmission Loss.

    Returns:
        float: The weighted sound reduction index.
    """
    return rw_curve(tl)[7]


def rw_c(tl: NDArray[np.float64]) -> float:
    """Calculate $R_W + C$ from a NumPy array `tl` with third octave data
    between 100 Hz and 3.15 kHz.

    Args:
        tl: Transmission Loss.

    Returns:
        float: The weighted sound reduction index with spectrum adaptation term C.
    """
    k = np.array(
        [-29, -26, -23, -21, -19, -17, -15, -13, -12, -11, -10, -9, -9, -9, -9, -9]
    )
    a = -10 * np.log10(np.sum(10 ** ((k - tl) / 10)))
    return a


def rw_ctr(tl: NDArray[np.float64]) -> float:
    """Calculate $R_W + C_{tr}$ from a NumPy array `tl` with third octave
    data between 100 Hz and 3.15 kHz.

    Args:
        tl: Transmission Loss.

    Returns:
        float: The weighted sound reduction index with spectrum adaptation term $C_{tr}$.
    """
    k_tr = np.array(
        [-20, -20, -18, -16, -15, -14, -13, -12, -11, -9, -8, -9, -10, -11, -13, -15]
    )
    a_tr = -10 * np.log10(np.sum(10 ** ((k_tr - tl) / 10)))
    return a_tr


def stc_curve(tl: NDArray[np.float64]) -> NDArray[np.float64]:
    """Calculate the Sound Transmission Class (STC) curve from a NumPy array `tl`
    with third octave data between 125 Hz and 4 kHz.

    Args:
        tl: Transmission Loss.

    Returns:
        NDArray[np.float64]: The STC reference curve values.
    """
    ref_curve = np.array([0, 3, 6, 9, 12, 15, 16, 17, 18, 19, 20, 20, 20, 20, 20, 20])
    top_curve = ref_curve
    res_sum = 0
    while True:
        diff = tl - top_curve
        residuals = np.clip(diff, np.min(diff), 0)
        res_sum = np.sum(residuals)
        if res_sum < -32:
            if np.any(residuals > -8):
                top_curve -= 1
                break
        top_curve += 1
    return top_curve


def stc(tl: NDArray[np.float64]) -> float:
    """Calculate the Sound Transmission Class (STC) from a NumPy array `tl` with
    third octave data between 125 Hz and 4 kHz.

    Args:
        tl: Transmission Loss.

    Returns:
        float: The Sound Transmission Class value.
    """
    return stc_curve(tl)[6]


def mass_law(
    freq: float | NDArray[np.float64],
    vol_density: float,
    thickness: float,
    theta: float = 0,
    c: float = 343,
    rho0: float = 1.225,
) -> float | NDArray[np.float64]:
    r"""Calculate transmission loss according to mass law.

    Args:
        freq: Frequency of interest in Hz.
        vol_density: Volumetric density of material in $frac{kg}{m^3}$.
        thickness: Thickness of wall in m.
        theta: Angle of incidence in degrees. Default value is 0 (normal incidence).
        c: Speed of sound in m/s. Defaults to 343.
        rho0: Density of air in $\frac{kg}{m^3}$. Defaults to 1.225.

    Returns:
        float | NDArray[np.float64]: Transmission loss value(s) in dB.
    """
    rad_freq = 2.0 * np.pi * freq
    surface_density = vol_density * thickness
    theta_rad = np.deg2rad(theta)
    a = rad_freq * surface_density * np.cos(theta_rad) / (2 * rho0 * c)
    tl_theta = 10 * np.log10(1 + a**2)
    return tl_theta


__all__ = ["rw_curve", "rw", "rw_c", "rw_ctr", "stc_curve", "stc", "mass_law"]



"""The cepstrum module contains functions for cepstral analysis of signals.

Functions:
  complex_cepstrum: Compute the complex cepstrum of a real sequence.
  real_cepstrum: Compute the real cepstrum of a real sequence.
  inverse_complex_cepstrum: Compute the inverse complex cepstrum of a real sequence.
  minimum_phase: Compute the minimum phase reconstruction of a real sequence.

Notes:
  The cepstrum is defined as the inverse Fourier transform of the logarithm of the
  Fourier transform of a signal. It has applications in speech processing, echo detection,
  and signal deconvolution.
"""

import numpy as np
from numpy.typing import NDArray
from typing import Tuple, Optional

__all__ = [
    "complex_cepstrum",
    "real_cepstrum",
    "inverse_complex_cepstrum",
    "minimum_phase",
]


def complex_cepstrum(
    x: NDArray[np.float64], n: Optional[int] = None
) -> Tuple[NDArray[np.float64], NDArray[np.int_]]:
    r"""Compute the complex cepstrum of a real sequence.

    The complex cepstrum is given by:
    $$
    c[n] = F^{-1}\left[\log_{10}\left(F{x[n]}\right)\right]
    $$

    where $x[n]$ is the input signal and $F$ and $F^{-1}$
    are respectively the forward and backward Fourier transform.


    Args:
      x: Real sequence to compute complex cepstrum of.
      n: Length of the Fourier transform.

    Returns:
      The complex cepstrum of the real data sequence `x` computed using the Fourier transform.
      The amount of samples of circular delay added to `x`.


    See Also:
      - [`real_cepstrum`][acoustic_toolbox.cepstrum.real_cepstrum]: Compute the real cepstrum.
      - [`inverse_complex_cepstrum`][acoustic_toolbox.cepstrum.inverse_complex_cepstrum]: Compute the inverse complex cepstrum of a real sequence.

    Examples:
      In the following example we use the cepstrum to determine the fundamental
      frequency of a set of harmonics. There is a distinct peak at the quefrency
      corresponding to the fundamental frequency. To be more precise, the peak
      corresponds to the spacing between the harmonics.

      >>> import numpy as np
      >>> import matplotlib.pyplot as plt
      >>> from acoustic_toolbox.cepstrum import complex_cepstrum

      >>> duration = 5.0
      >>> fs = 8000.0
      >>> samples = int(fs*duration)
      >>> t = np.arange(samples) / fs

      >>> fundamental = 100.0
      >>> harmonics = np.arange(1, 30) * fundamental
      >>> signal = np.sin(2.0*np.pi*harmonics[:,None]*t).sum(axis=0)
      >>> ceps, _ = complex_cepstrum(signal)

      >>> fig = plt.figure()
      >>> ax0 = fig.add_subplot(211)
      >>> ax0.plot(t, signal)
      >>> ax0.set_xlabel('time in seconds')
      >>> ax0.set_xlim(0.0, 0.05)
      >>> ax1 = fig.add_subplot(212)
      >>> ax1.plot(t, ceps)
      >>> ax1.set_xlabel('quefrency in seconds')
      >>> ax1.set_xlim(0.005, 0.015)
      >>> ax1.set_ylim(-5., +10.)

    References:
      1. Wikipedia, "Cepstrum".
            [http://en.wikipedia.org/wiki/Cepstrum](http://en.wikipedia.org/wiki/Cepstrum)
      2. M.P. Norton and D.G. Karczub, D.G.,
            "Fundamentals of Noise and Vibration Analysis for Engineers", 2003.
      3. B. P. Bogert, M. J. R. Healy, and J. W. Tukey:
            "The Quefrency Analysis of Time Series for Echoes: Cepstrum, Pseudo
            Autocovariance, Cross-Cepstrum and Saphe Cracking".
            Proceedings of the Symposium on Time Series Analysis
            Chapter 15, 209-243. New York: Wiley, 1963.
    """

    def _unwrap(
        phase: NDArray[np.float64],
    ) -> Tuple[NDArray[np.float64], NDArray[np.int_]]:
        """Unwrap phase values.

        Args:
          phase: Phase values to unwrap.

        Returns:
          Unwrapped phase values
          Number of delay samples
        """
        samples = phase.shape[-1]
        unwrapped = np.unwrap(phase)
        center = (samples + 1) // 2
        if samples == 1:
            center = 0
        ndelay = np.array(np.round(unwrapped[..., center] / np.pi))
        unwrapped -= np.pi * ndelay[..., None] * np.arange(samples) / center
        return unwrapped, ndelay

    spectrum = np.fft.fft(x, n=n)
    unwrapped_phase, ndelay = _unwrap(np.angle(spectrum))
    log_spectrum = np.log(np.abs(spectrum)) + 1j * unwrapped_phase
    ceps = np.fft.ifft(log_spectrum).real

    return ceps, ndelay


def real_cepstrum(
    x: NDArray[np.float64], n: Optional[int] = None
) -> NDArray[np.float64]:
    r"""Compute the real cepstrum of a real sequence.

    The real cepstrum is given by:
    $$
    c[n] = F^{-1}\left[\log_{10}\left|F{x[n]}\right|\right]
    $$

    where $x[n]$ is the input signal and $F$ and $F^{-1}$ are respectively
    the forward and backward Fourier transform.

    Note that contrary to the complex cepstrum the magnitude is taken of the spectrum.

    Args:
      x: Real sequence to compute real cepstrum of.
      n: Length of the Fourier transform.

    Returns:
      The real cepstrum.


    See Also:
      - [`complex_cepstrum`][acoustic_toolbox.cepstrum.complex_cepstrum]: Compute the complex cepstrum of a real sequence.
      - [`inverse_complex_cepstrum`][acoustic_toolbox.cepstrum.inverse_complex_cepstrum]: Compute the inverse complex cepstrum of a real sequence.

    Examples:
      >>> from acoustic_toolbox.cepstrum import real_cepstrum

    References:
      1. Wikipedia, "Cepstrum".
          [http://en.wikipedia.org/wiki/Cepstrum](http://en.wikipedia.org/wiki/Cepstrum)
    """
    spectrum = np.fft.fft(x, n=n)
    ceps = np.fft.ifft(np.log(np.abs(spectrum))).real

    return ceps


def inverse_complex_cepstrum(
    ceps: NDArray[np.float64], ndelay: NDArray[np.int_]
) -> NDArray[np.float64]:
    r"""Compute the inverse complex cepstrum of a real sequence.

    The inverse complex cepstrum is given by:
    $$
    x[n] = F^{-1}\left[\exp(F(c[n]))\right]
    $$

    where $c[n]$ is the input signal and $F$ and $F^{-1}$ are respectively the forward and backward Fourier transform.

    Args:
      ceps: Real sequence to compute inverse complex cepstrum of.
      ndelay: The amount of samples of circular delay added to `x`.

    Returns:
      The inverse complex cepstrum of the real sequence `ceps`.

    See Also:
      - [`complex_cepstrum`][acoustic_toolbox.cepstrum.complex_cepstrum]: Compute the complex cepstrum of a real sequence.
      - [`real_cepstrum`][acoustic_toolbox.cepstrum.real_cepstrum]: Compute the real cepstrum of a real sequence.

    Examples:
      Taking the complex cepstrum and then the inverse complex cepstrum results
      in the original sequence.

      >>> import numpy as np
      >>> from acoustic_toolbox.cepstrum import inverse_complex_cepstrum
      >>> x = np.arange(10)
      >>> ceps, ndelay = complex_cepstrum(x)
      >>> y = inverse_complex_cepstrum(ceps, ndelay)
      >>> print(x)
      >>> print(y)

    References:
      1. Wikipedia, "Cepstrum".
          [http://en.wikipedia.org/wiki/Cepstrum](http://en.wikipedia.org/wiki/Cepstrum)
    """

    def _wrap(
        phase: NDArray[np.float64], ndelay: NDArray[np.int_]
    ) -> NDArray[np.float64]:
        """Wrap phase values.

        Args:
          phase: Phase values to wrap.
          ndelay: Number of delay samples.

        Returns:
          Wrapped phase values.
        """
        ndelay = np.array(ndelay)
        samples = phase.shape[-1]
        center = (samples + 1) // 2
        wrapped = phase + np.pi * ndelay[..., None] * np.arange(samples) / center
        return wrapped

    log_spectrum = np.fft.fft(ceps)
    spectrum = np.exp(log_spectrum.real + 1j * _wrap(log_spectrum.imag, ndelay))
    x = np.fft.ifft(spectrum).real
    return x


def minimum_phase(
    x: NDArray[np.float64], n: Optional[int] = None
) -> NDArray[np.float64]:
    """Compute the minimum phase reconstruction of a real sequence.

    Args:
      x: Real sequence to compute the minimum phase reconstruction of.
      n: Length of the Fourier transform.

    Returns:
      The minimum phase reconstruction of the real sequence `x`.

    Compute the minimum phase reconstruction of a real sequence using the real cepstrum.

    See Also:
      - [`real_cepstrum`][acoustic_toolbox.cepstrum.real_cepstrum]: Compute the real cepstrum.

    Examples:
      >>> from acoustic_toolbox.cepstrum import minimum_phase

    References:
      1. Soo-Chang Pei, Huei-Shan Lin. Minimum-Phase FIR Filter Design Using
          Real Cepstrum. IEEE TRANSACTIONS ON CIRCUITS AND SYSTEMS-II:
          EXPRESS BRIEFS, VOL. 53, NO. 10, OCTOBER 2006
    """
    if n is None:
        n = len(x)
    ceps = real_cepstrum(x, n=n)
    odd = n % 2
    window = np.concatenate(
        (
            [1.0],
            2.0 * np.ones((n + odd) // 2 - 1),
            np.ones(1 - odd),
            np.zeros((n + odd) // 2 - 1),
        )
    )

    m = np.fft.ifft(np.exp(np.fft.fft(window * ceps))).real

    return m



import numpy as np

from acoustic_toolbox.standards.iec_61672_1_2013 import time_weighted_level, time_averaged_level


def signal_fs():
    fs = 4000.0
    f = 400.0
    duration = 3.0
    samples = int(duration * fs)
    t = np.arange(samples) / fs
    x = np.sin(2.0 * np.pi * f * t)
    return x, fs


def test_fast_level():
    r"""Test whether integration with time-constant FAST gives the correct level.

    Note that the reference sound pressure is used.

    In this test the amplitude of the sine is 1, which means the mean squared $MS$ is 0.5
    With a reference pressure $p_r$ of 2.0e-5 the level should be 91 decibel

    $$
    L = 10 \cdot \log_{10}{\left(\frac{MS}{p_r^2} \right)}
    $$

    $$
    L = 10 \cdot \log_{10}{\left(\frac{0.5}{(2e-5)^2} \right)} = 91
    $$
    """
    x, fs = signal_fs()

    times, levels = time_weighted_level(x, fs, time_mode="fast")
    assert abs(levels[-1] - 91) < 0.05

    x *= 4.0
    times, levels = time_weighted_level(x, fs, time_mode="fast")
    assert abs(levels[-1] - 103) < 0.05

def test_slow_level():
    """Test whether integration with time-constant SLOW gives the correct level."""
    x, fs = signal_fs()
    x = np.concatenate([x, x]) # Make signal longer for SLOW weighting

    times, levels = time_weighted_level(x, fs, time_mode="slow")
    assert abs(levels[-1] - 91) < 0.05

    x *= 4.0
    times, levels = time_weighted_level(x, fs, time_mode="slow")
    assert abs(levels[-1] - 103) < 0.05


def test_time_averaged_level():
    """Test whether time-averaged level gives the correct level."""
    x, fs = signal_fs()

    # Test with integration time of 1.0 second
    integration_time = 1.0
    times, levels = time_averaged_level(x, fs, integration_time)

    # Expected level for sine wave with amp 1 is approx 91 dB relative to 2e-5
    # 10 * log10(0.5 / (2e-5)**2) = 90.969...
    assert abs(levels[-1] - 91) < 0.05

    # Test with scaling
    x_scaled = x * 4.0
    times, levels = time_averaged_level(x_scaled, fs, integration_time)
    # 20 * log10(4) = 12.04 dB increase
    assert abs(levels[-1] - 103) < 0.05

    # Test with fractional integration time to exercise the variable chunking
    # fs = 4000. integration_time 0.1001s -> 400.4 samples.
    integration_time = 0.1001
    times, levels = time_averaged_level(x, fs, integration_time)

    # Should still be approximately correct
    assert abs(np.mean(levels) - 91) < 0.5


def test_frequency_weighting_A():
    """Test A-weighting frequency response."""
    from acoustic_toolbox.standards.iec_61672_1_2013 import (
        frequency_weighting,
        NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES,
        WEIGHTING_A,
    )

    fs = 48000
    duration = 0.5
    t = np.arange(int(fs * duration)) / fs

    # Tolerances (dB)
    tolerance = 0.6

    for f_nom, expected_db in zip(NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES, WEIGHTING_A):
        if f_nom < 10:
            continue

        if f_nom > 8000:
            # Due to using bilinear transform, the accuracy degrades near Nyquist so let's not test those
            # other techniques (e.g., improved matched z-transform) would be needed for higher accuracy
            continue

        # Generate sine wave at this frequency
        signal = np.sin(2 * np.pi * f_nom * t)

        # Apply weighting
        weighted = frequency_weighting(signal, fs, weighting='A')

        # Calculate RMS levels - use last part to avoid transient
        valid_idx = int(fs * 0.2)
        valid_part = weighted[valid_idx:]
        rms_in = np.sqrt(np.mean(signal[valid_idx:]**2))
        rms_out = np.sqrt(np.mean(valid_part**2))

        db_in = 20 * np.log10(rms_in)
        db_out = 20 * np.log10(rms_out)

        attenuation = db_out - db_in

        # Check if attenuation matches expected A-weighting
        assert abs(attenuation - expected_db) < tolerance, \
            f"Frequency {f_nom} Hz: Expected {expected_db} dB, got {attenuation:.2f} dB"


def test_frequency_weighting_C():
    """Test C-weighting frequency response."""
    from acoustic_toolbox.standards.iec_61672_1_2013 import (
        frequency_weighting,
        NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES,
        WEIGHTING_C,
    )

    fs = 48000
    duration = 0.5
    t = np.arange(int(fs * duration)) / fs
    # Tolerances (dB)
    tolerance = 0.6

    for f_nom, expected_db in zip(NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES, WEIGHTING_C):
        if f_nom < 10:
            continue

        if f_nom > 8000:
            # Due to using bilinear transform, the accuracy degrades near Nyquist so let's not test those
            # other techniques (e.g., improved matched z-transform?) would be needed for higher accuracy
            continue

        # Generate sine wave at this frequency
        signal = np.sin(2 * np.pi * f_nom * t)

        # Apply weighting
        weighted = frequency_weighting(signal, fs, weighting='C')

        # Calculate RMS levels - use last part to avoid transient
        valid_idx = int(fs * 0.2)
        valid_part = weighted[valid_idx:]
        rms_in = np.sqrt(np.mean(signal[valid_idx:]**2))
        rms_out = np.sqrt(np.mean(valid_part**2))

        db_in = 20 * np.log10(rms_in)
        db_out = 20 * np.log10(rms_out)

        attenuation = db_out - db_in

        # Check if attenuation matches expected C-weighting
        assert abs(attenuation - expected_db) < tolerance, \
            f"Frequency {f_nom} Hz: Expected {expected_db} dB, got {attenuation:.2f} dB"


def test_frequency_weighting_Z():
    """Test Z-weighting frequency response (should be flat)."""
    from acoustic_toolbox.standards.iec_61672_1_2013 import frequency_weighting

    fs = 48000
    signal = np.random.randn(fs) # 1 second of noise

    weighted = frequency_weighting(signal, fs, weighting='Z')

    # Z-weighting should return the signal as-is
    np.testing.assert_array_equal(signal, weighted)


def test_frequency_weighting_zero_phase():
    """Test zero-phase filtering."""
    from acoustic_toolbox.standards.iec_61672_1_2013 import frequency_weighting

    fs = 1000
    # Create an impulse
    signal = np.zeros(1000)
    signal[500] = 1.0

    # Apply A-weighting with zero phase
    weighted = frequency_weighting(signal, fs, weighting='A', zero_phase=True)

    # Check for symmetry around the impulse to confirm zero phase behavior
    # The impulse response should be symmetric
    max_idx = np.argmax(np.abs(weighted))
    assert max_idx == 500, "Peak should remain at center"

    # Check symmetry of a small window around peak
    window = 10
    left = weighted[max_idx-window:max_idx]
    right = weighted[max_idx+1:max_idx+window+1][::-1]

    np.testing.assert_allclose(left, right, atol=1e-10)



"""ISO 1996-1:2003 defines the basic quantities to be used for the description of
noise in community environments and describes basic assessment procedures. It
also specifies methods to assess environmental noise and gives guidance on
predicting the potential annoyance response of a community to long-term exposure
from various types of environmental noises. The sound sources can be separate or
in various combinations. Application of the method to predict annoyance response
is limited to areas where people reside and to related long-term land uses.

Reference:
    ISO 1996-1:2003: Description, measurement and assessment of environmental noise
"""

import numpy as np


def composite_rating_level(
    levels: np.ndarray, hours: np.ndarray, adjustment: np.ndarray
) -> float | np.ndarray:
    r"""Composite rating level.

    The composite rating level is calculated as:

    $$
    L_R = 10 \log{\left[ \sum_i \frac{d_i}{24} 10^{(L_i+K_i)/10}  \right]}
    $$

    where $i$ is a period. See equation 6 and 7 of the standard.

    Note:
      Summation is done over the last axis

    Args:
        levels: Level per period
        hours: Amount of hours per period
        adjustment: Adjustment per period

    Returns:
        The composite rating level in dB as a float if a scalar result is obtained, or as
        an ndarray if multiple periods are processed.

    """
    levels = np.asarray(levels)
    hours = np.asarray(hours)
    adjustment = np.asarray(adjustment)

    return 10.0 * np.log10(
        (hours / 24.0 * 10.0 ** ((levels + adjustment) / 10.0)).sum(axis=-1)
    )



"""ISO/TR 25417:2007 specifies definitions of acoustical quantities and terms used
in noise measurement documents prepared by ISO Technical Committee TC 43,
Acoustics, Subcommittee SC 1, Noise, together with their symbols and units, with
the principal aim of harmonizing the terminology used in [ISO/TR 25417](http://www.iso.org/iso/home/store/catalogue_tc/catalogue_detail.htm?csnumber=42915).
"""  # noqa: D205

import numpy as np

REFERENCE_PRESSURE = 2.0e-5
r"""Reference value of the sound pressure $p_0$ is $2 \cdot 10^{-5}$ Pa."""


def sound_pressure_level(
    pressure, reference_pressure=REFERENCE_PRESSURE
) -> float | np.ndarray:
    r"""Sound pressure level $L_p$ in dB.

    The sound pressure level is calculated as:
    $$
    L_p = 10 \log_{10}{ \left( \frac{p^2}{p_0^2} \right)}
    $$

    See section 2.2 of the standard.

    Args:
        pressure: Instantaneous sound pressure $p$.
        reference_pressure: Reference value $p_0$.

    Returns:
        Sound pressure level
    """
    return 10.0 * np.log10(pressure**2.0 / reference_pressure**2.0)


def equivalent_sound_pressure_level(
    pressure, reference_pressure=REFERENCE_PRESSURE, axis=-1
) -> float | np.ndarray:
    r"""Time-averaged sound pressure level $L_{p,T}$ or equivalent-continious sound pressure level $L_{p,eqT}$ in dB.

    The time-averaged sound pressure level is calculated as:
    $$
    L_{p,T} = L_{p,eqT} = 10.0 \log_{10}{ \left( \frac{\frac{1}{T} \int_{t_1}^{t_2} p^2 (t) \mathrm{d} t  }{p_0^2} \right)}
    $$

    See section 2.3 of the standard.

    Args:
        pressure: Instantaneous sound pressure $p$.
        reference_pressure: Reference value $p_0$.
        axis: Axis

    Returns:
        Time-averaged sound pressure level
    """
    return 10.0 * np.log10((pressure**2.0).mean(axis=axis) / reference_pressure**2.0)


def max_sound_pressure_level(
    pressure, reference_pressure=REFERENCE_PRESSURE, axis=-1
) -> float | np.ndarray:
    r"""Maximum time-averaged sound pressure level $L_{F,max}$ in dB.

    Args:
        pressure: Instantaneous sound pressure $p$.
        reference_pressure: Reference value $p_0$.
        axis: Axis

    Returns:
        Maximum sound pressure level $\mathrm{max}{(L_{p})}$
    """
    return sound_pressure_level(pressure, reference_pressure=reference_pressure).max(
        axis=axis
    )


def peak_sound_pressure(pressure, axis=-1) -> float | np.ndarray:
    r"""Peak sound pressure $p_{peak}$ is the greatest absolute sound pressure during a certain time interval.

    Args:
        pressure: Instantaneous sound pressure $p$.
        axis: Axis

    Returns:
        Peak sound pressure $p_{peak} = \mathrm{max}(|p|)$
    """
    return np.abs(pressure).max(axis=axis)


def peak_sound_pressure_level(
    pressure, reference_pressure=REFERENCE_PRESSURE, axis=-1
) -> float | np.ndarray:
    r"""Peak sound pressure level $L_{p,peak}$ in dB.

    Args:
        pressure: Instantaneous sound pressure $p$.
        reference_pressure: Reference value $p_0$.
        axis: Axis

    Returns:
        Peak sound pressure level
            $$
            L_{p,peak} = 10.0 \log \frac{p_{peak}^2.0}{p_0^2}
            $$
    """
    return 10.0 * np.log10(
        peak_sound_pressure(pressure, axis=axis) ** 2.0 / reference_pressure**2.0
    )


REFERENCE_SOUND_EXPOSURE = 4.0e-10
r"""
Reference value of the sound exposure $E_0$ is $4 \cdot 10^{-12} \mathrm{Pa}^2\mathrm{s}$.
"""


def sound_exposure(pressure, fs, axis=-1) -> float | np.ndarray:
    r"""Calculate sound exposure $E_T$.

    Args:
        pressure: Instantaneous sound pressure $p$.
        fs: Sample frequency $f_s$.
        axis: Axis

    Returns:
        Sound exposure
            $$
            E_T = \int_{t_1}^{t_2} p^2(t) \mathrm{d}t
            $$
    """
    return (pressure**2.0 / fs).sum(axis=axis)


def sound_exposure_level(
    pressure, fs, reference_sound_exposure=REFERENCE_SOUND_EXPOSURE, axis=-1
) -> float | np.ndarray:
    r"""Sound exposure level $L_{E,T}$ in dB.

    Args:
        pressure: Instantaneous sound pressure $p$.
        fs: Sample frequency $f_s$.
        reference_sound_exposure: Reference sound exposure $E_T$.
        axis: Axis

    Returns:
        Sound exposure level
            $$
            L_{E,T} = 10 \log_{10}{ \frac{E_T}{E_0}  }
            $$
    """
    return 10.0 * np.log10(
        sound_exposure(pressure, fs, axis=axis) / reference_sound_exposure
    )


REFERENCE_POWER = 1.0e-12
"""
Reference value of the sound power $P_0$ is 1 pW.
"""


def sound_power_level(power, reference_power=REFERENCE_POWER) -> float | np.ndarray:
    r"""Sound power level $L_W$.

    Args:
        power: Sound power $P$.
        reference_power: Reference sound power $P_0$.

    Returns:
        Sound power level calculated as:
            $$
            10 \log_{10}{ \frac{P}{P_0}  }
            $$
    """
    return 10.0 * np.log10(power / reference_power)


def sound_energy(power, axis=-1) -> float | np.ndarray:
    r"""Sound energy $J$.

    Args:
        power: Sound power $P$.
        axis: Axis

    Returns:
        Sound energy
            $$
            J = \int_{t_1}^{t_2} P(t) \mathrm{d} t
            $$
    """
    return power.sum(axis=axis)


REFERENCE_ENERGY = 1.0e-12
"""Reference value of the sound energy $J_0$ is 1 pJ."""


def sound_energy_level(energy, reference_energy=REFERENCE_ENERGY) -> np.ndarray:
    r"""Sound energy level $L_J$ in dB.

    Args:
        energy: Sound energy $J$.
        reference_energy: Reference sound energy $J_0$.

    Returns:
        Sound energy level
            $$
            L_{J} = 10 \log_{10}{ \frac{J}{J_0} }
            $$
    """
    return np.log10(energy / reference_energy)


def sound_intensity(pressure, velocity) -> np.ndarray:
    r"""Sound intensity $\mathbf{i}$.

    Args:
        pressure: Sound pressure $p(t)$.
        velocity: Particle velocity $\mathbf{u}(t)$.

    Returns:
        Sound intensity
            $$
            \mathbf{i} = p(t) \cdot \mathbf{u}(t)
            $$
    """
    return pressure * velocity


REFERENCE_INTENSITY = 1.0e-12
r"""Reference value of the sound intensity $I_0$ is $\mathrm{1 pW/m^2}$."""


def time_averaged_sound_intensity(intensity, axis=-1) -> float:
    r"""Time-averaged sound intensity $\mathbf{I}_T$.

    Args:
        intensity: Sound intensity $\mathbf{i}$.
        axis: Axis

    Returns:
        Time-averaged sound intensity
            $$
            I_T = \frac{1}{T} \int_{t_1}^{t_2} \mathbf{i}(t)
            $$
    """
    return intensity.mean(axis=axis)


def time_averaged_sound_intensity_level(
    time_averaged_sound_intensity, reference_intensity=REFERENCE_INTENSITY, axis=-1
) -> float:
    r"""Time-averaged sound intensity level $L_{I,T}$.

    Args:
        time_averaged_sound_intensity: Time-averaged sound intensity $\mathbf{I}_T$.
        reference_intensity: Reference sound intensity $I_0$.
        axis: Axis along which to calculate norm.

    Returns:
        Time-averaged sound intensity level calculated as:
            $$
            L_{I,T} = 10 \log_{10} { \frac{|\mathbf{I}_T|}{I_0} }
            $$
    """
    return 10.0 * np.log10(
        np.linalg.norm(time_averaged_sound_intensity, axis=axis) / reference_intensity
    )


def normal_time_averaged_sound_intensity(
    time_averaged_sound_intensity, unit_normal_vector
) -> float:
    r"""Normal time-averaged sound intensity $\mathbf{I}_{n,T}$.

    Args:
        time_averaged_sound_intensity: Time-averaged sound intensity $\mathbf{I}_T$.
        unit_normal_vector: Unit normal vector $\mathbf{n}$.

    Returns:
        Normal time-averaged sound intensity
            $$
            I_{n,T} = \mathbf{I}_T \cdot \mathbf{n}
            $$
    """
    return time_averaged_sound_intensity.dot(unit_normal_vector)


def normal_time_averaged_sound_intensity_level(
    normal_time_averaged_sound_intensity, reference_intensity=REFERENCE_INTENSITY
) -> float:
    r"""Normal time-averaged sound intensity level $L_{I_{n,T}}$ in dB.

    Args:
        normal_time_averaged_sound_intensity: Normal time-averaged sound intensity $I_{n,T}$.
        reference_intensity: Reference sound intensity $I_0$.

    Returns:
        Normal time-averaged sound intensity level calculated as:
            $$
            I_{n,T} = 10 \log_{10} { \frac{|I_{n,T}|}{I_0}}
            $$
    """
    return 10.0 * np.log10(
        np.abs(normal_time_averaged_sound_intensity / reference_intensity)
    )



import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_array_equal,
    assert_array_almost_equal,
)

from acoustic_toolbox.utils import esum, _is_1d, mean_tl, wavelength, w


class TestUtils:
    """Test :mod:`acoustic_toolbox.utils`"""

    def test_mean_tl_1d(self):
        tl = np.array([20, 30, 40, 50])
        surfaces = np.array([50, 40, 30, 20])
        tl_real = 24.1116827
        calculated = mean_tl(tl, surfaces)
        assert_almost_equal(calculated, tl_real)

    def test_mean_tl_2d(self):
        tl = np.array([[20, 30, 40, 50], [20, 30, 40, 50]])
        surfaces = np.array([[50, 40, 30, 20], [1, 10, 11, 22]])
        tl_real = np.array([24.1116827, 33.1466548])
        calculated = mean_tl(tl, surfaces)
        assert_array_almost_equal(calculated, tl_real)

    def test_esum_1d(self):
        calculated = esum(np.array([90, 90, 90]))
        real = 94.77121255
        assert_almost_equal(calculated, real)

    def test_esum_2d_default_axis(self):
        calculated = esum(np.array([[90, 90, 90], [80, 80, 80]]))
        real = np.array(95.18513939877889)
        # real = np.array([94.77121255, 84.77121255])
        assert_array_almost_equal(calculated, real)

    def test_esum_2d_axis0(self):
        calculated = esum(np.array([[90, 90, 90], [80, 80, 80]]), axis=0)
        real = np.array([90.41392685, 90.41392685, 90.41392685])
        assert_almost_equal(calculated, real)

    def test__is_1d_float(self):
        a = 0.9
        is_float = _is_1d(a)
        assert a == is_float

    def test__is_1d_1darray(self):
        a = np.array([0.1, 0.2, 0.3])
        is_1d_array = _is_1d(a)
        a_return = np.array([a])
        assert_array_equal(a_return, is_1d_array)

    def test__is_1d_2darray(self):
        a = np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])
        is_2d_array = _is_1d(a)
        assert_array_equal(a, is_2d_array)

    def test__is1d_vector_2darray(self):
        a = np.array([[0.1, 0.2, 0.3]])
        is_vector_2darray = _is_1d(a)
        a_return = np.array([0.1, 0.2, 0.3])
        assert_array_equal(a_return, is_vector_2darray)

    def test_wavelength1(self):
        calculated = wavelength(200)
        quasi = 1.715
        assert_almost_equal(calculated, quasi)

    def test_wavelength2(self):
        calculated = wavelength(200, 343)
        quasi = 1.715
        assert_almost_equal(calculated, quasi)

    def test_wavelength3(self):
        freqs = np.array([555, 3333, 11111])
        calculated = wavelength(freqs)
        quasi = np.array([0.618018018, 0.102910291, 0.030870308])
        assert_array_almost_equal(calculated, quasi)

    def test_w1(self):
        calculated = w(1000)
        quasi = 6283.185307179
        assert_almost_equal(calculated, quasi)

    def test_w2(self):
        calculated = w(np.array([1000, 2000, 3000]))
        quasi = np.array([6283.185307179, 12566.370614359, 18849.555921538])
        assert_array_almost_equal(calculated, quasi)



"""Standards module for acoustic calculations.

This module provides implementations of various acoustic standards including:

- ISO TR 25417:2007 - Acoustical quantity definitions
- IEC 61672-1:2013 - Sound level meters
- IEC 61260-1:2014 - Octave-band filters
- ISO 9613-1:1993 - Atmospheric sound attenuation
- ISO 1996-1:2003 - Environmental noise assessment
- ISO 1996-2:2007 - Environmental noise level determination
"""

from acoustic_toolbox.standards import (
    iso_tr_25417_2007,
    iec_61672_1_2013,
    iec_61260_1_2014,
    iso_9613_1_1993,
    iso_1996_1_2003,
    iso_1996_2_2007,
)

__all__ = [
    "iso_tr_25417_2007",
    "iec_61672_1_2013",
    "iec_61260_1_2014",
    "iso_9613_1_1993",
    "iso_1996_1_2003",
    "iso_1996_2_2007",
]
